# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest
from botocore.session import Session
from prompt_toolkit.application import Application
from prompt_toolkit.completion import Completion, PathCompleter
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import walk

from awscli.customizations.wizard.app import (
    FileIO,
    WizardAppRunner,
    WizardTraverser,
    WizardValues,
)
from awscli.customizations.wizard.core import BaseStep, Executor
from awscli.customizations.wizard.exceptions import (
    InvalidChoiceException,
    InvalidDataTypeConversionException,
    UnableToRunWizardError,
    UnexpectedWizardException,
)
from awscli.customizations.wizard.factory import create_wizard_app
from awscli.testutils import mock, unittest
from tests import PromptToolkitAppRunner


@pytest.fixture
def mock_botocore_session():
    return mock.Mock(spec=Session)


@pytest.fixture
def mock_iam_client(mock_botocore_session):
    mock_client = mock.Mock()
    mock_botocore_session.create_client.return_value = mock_client
    mock_client.list_policies.return_value = {
        'Policies': [
            {'PolicyName': 'policy1', 'Arn': 'policy1_arn'},
            {'PolicyName': 'policy2', 'Arn': 'policy2_arn'},
        ]
    }
    mock_client.get_policy.return_value = {
        'Policy': {'DefaultVersionId': 'policy_id'}
    }
    mock_client.get_policy_version.return_value = {
        'PolicyVersion': {'Document': {'policy': 'policy_document'}}
    }
    mock_client.create_role.return_value = {
        'Role': {'Arn': 'returned-role-arn'}
    }
    return mock_client


@pytest.fixture
def make_stubbed_wizard_runner(ptk_app_session, mock_botocore_session):
    def _make_stubbed_wizard_runner(definition):
        app = create_wizard_app(
            definition=definition,
            session=mock_botocore_session,
            output=ptk_app_session.output,
            app_input=ptk_app_session.input,
        )
        ptk_app_session.app = app
        return PromptToolkitAppRunner(app=app)

    yield _make_stubbed_wizard_runner


@pytest.fixture
def patch_path_completer():
    with mock.patch(
        'awscli.customizations.wizard.ui.prompt.PathCompleter',
        FakePathCompleter,
    ) as completer:
        yield completer


class FakePathCompleter(PathCompleter):
    def get_completions(self, document, complete_event):
        prefix_len = len(document.text)
        yield from [
            Completion('file1'[prefix_len:], 0, display='file1'),
            Completion('file2'[prefix_len:], 0, display='file2'),
        ]


@pytest.fixture
def empty_definition():
    return {
        'title': 'Wizard title',
        'plan': {
            '__DONE__': {},
        },
        'execute': {},
    }


@pytest.fixture
def basic_definition():
    return {
        'title': 'Basic Wizard',
        'plan': {
            'first_section': {
                'shortname': 'First',
                'values': {
                    'prompt1': {
                        'description': 'Description of first prompt',
                        'type': 'prompt',
                    },
                    'prompt2': {
                        'description': 'Description of second prompt',
                        'type': 'prompt',
                        'default_value': 'foo',
                    },
                },
            },
            'second_section': {
                'shortname': 'Second',
                'values': {
                    'second_section_prompt': {
                        'description': 'Description of prompt',
                        'type': 'prompt',
                    },
                    'template_section': {
                        'type': 'template',
                        'value': 'some text',
                    },
                },
            },
            '__DONE__': {},
        },
        'execute': {},
        '__OUTPUT__': {'value': 'output: {template_section}'},
    }


@pytest.fixture
def conditional_definition():
    return {
        'title': 'Conditional wizard',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'before_conditional': {
                        'description': 'Description of first prompt',
                        'type': 'prompt',
                    },
                    'conditional': {
                        'description': 'Description of second prompt',
                        'type': 'prompt',
                        'condition': {
                            'variable': 'before_conditional',
                            'equals': 'condition-met',
                        },
                    },
                    'after_conditional': {
                        'description': 'Description of second prompt',
                        'type': 'prompt',
                    },
                },
            },
            '__DONE__': {},
        },
        'execute': {},
    }


@pytest.fixture
def choices_definition():
    return {
        'title': 'Conditional wizard',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'choices_prompt': {
                        'description': 'Description of first prompt',
                        'type': 'prompt',
                        'choices': [
                            {
                                'display': 'Option 1',
                                'actual_value': 'actual_option_1',
                            },
                            {
                                'display': 'Option 2',
                                'actual_value': 'actual_option_2',
                            },
                        ],
                    }
                },
            },
            '__DONE__': {},
        },
        'execute': {},
    }


@pytest.fixture
def corrupted_choice_definition():
    return {
        'title': 'Conditional wizard',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'choices_prompt': {
                        'description': 'Description of first prompt',
                        'type': 'prompt',
                        'choices': 'corrupted_choice',
                    },
                    'corrupted_choice': {
                        'description': 'Description of first prompt',
                        'type': 'apicall',
                    },
                },
            },
            '__DONE__': {},
        },
        'execute': {},
    }


@pytest.fixture
def mixed_prompt_definition():
    return {
        'title': 'Wizard with both buffer inputs and select inputs',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'buffer_input_prompt': {
                        'description': 'Type answer',
                        'type': 'prompt',
                    },
                    'select_prompt': {
                        'description': 'Select answer',
                        'type': 'prompt',
                        'choices': ['select_answer_1', 'select_answer_2'],
                    },
                },
            },
            '__DONE__': {},
        },
        'execute': {},
    }


@pytest.fixture
def data_convert_definition():
    return {
        'title': 'Wizard with both buffer inputs and select inputs',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'buffer_input_int': {
                        'description': 'Type answer',
                        'type': 'prompt',
                        'datatype': 'int',
                    },
                    'buffer_input_bool': {
                        'description': 'Type answer',
                        'type': 'prompt',
                        'datatype': 'bool',
                    },
                },
            },
            '__DONE__': {},
        },
        'execute': {},
    }


@pytest.fixture
def api_call_definition():
    return {
        'title': 'Uses API call in prompting stage',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'existing_policies': {
                        'type': 'apicall',
                        'operation': 'iam.ListPolicies',
                        'params': {},
                        'query': (
                            'sort_by(Policies[].{display: PolicyName, '
                            'actual_value: Arn}, &display)'
                        ),
                    },
                    'choose_policy': {
                        'description': 'Choose policy',
                        'type': 'prompt',
                        'choices': 'existing_policies',
                    },
                },
            },
            '__DONE__': {},
        },
        'execute': {},
    }


@pytest.fixture
def details_definition():
    return {
        'title': 'Show details in prompting stage',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'version_id': {
                        'type': 'apicall',
                        'operation': 'iam.GetPolicy',
                        'params': {'PolicyArn': '{policy_arn}'},
                        'query': 'Policy.DefaultVersionId',
                        'cache': True,
                    },
                    'policy_document': {
                        'type': 'apicall',
                        'operation': 'iam.GetPolicyVersion',
                        'params': {
                            'PolicyArn': '{policy_arn}',
                            'VersionId': '{version_id}',
                        },
                        'query': 'PolicyVersion.Document',
                        'cache': True,
                    },
                    'existing_policies': {
                        'type': 'apicall',
                        'operation': 'iam.ListPolicies',
                        'params': {},
                        'query': (
                            'sort_by(Policies[].{display: PolicyName, '
                            'actual_value: Arn}, &display)'
                        ),
                    },
                    'policy_arn': {
                        'description': 'Choose policy',
                        'type': 'prompt',
                        'choices': 'existing_policies',
                        'details': {
                            'value': 'policy_document',
                            'description': 'Policy Document',
                            'output': 'json',
                        },
                    },
                    'some_prompt': {
                        'description': 'Choose something',
                        'type': 'prompt',
                        'choices': [1, 2, 3],
                    },
                },
            },
            '__DONE__': {},
        },
        'execute': {},
    }


@pytest.fixture
def preview_definition():
    return {
        'title': 'Show details in prompting stage',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'option1': {
                        'type': 'template',
                        'value': "First option details",
                    },
                    'option2': {
                        'type': 'template',
                        'value': "Second option details",
                    },
                    'some_prompt': {
                        'description': 'Choose something',
                        'type': 'prompt',
                        'choices': [{'display': '1', 'actual_value': '2'}],
                    },
                    'choose_option': {
                        'description': 'Choose option',
                        'type': 'prompt',
                        'choices': [
                            {'display': 'Option 1', 'actual_value': 'option1'},
                            {'display': 'Option 2', 'actual_value': 'option2'},
                        ],
                        'details': {
                            'visible': True,
                            'value': '__selected_choice__',
                            'description': 'Option details',
                        },
                    },
                },
            },
            '__DONE__': {},
        },
    }


@pytest.fixture
def shared_config_definition():
    return {
        'title': 'Uses shared config API in prompting stage',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'existing_profiles': {
                        'type': 'sharedconfig',
                        'operation': 'ListProfiles',
                    },
                    'choose_profile': {
                        'description': 'Choose profile',
                        'type': 'prompt',
                        'choices': 'existing_profiles',
                    },
                },
            },
            '__DONE__': {},
        },
    }


@pytest.fixture
def run_wizard_definition():
    return {
        'title': 'For running execute step',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'role_name': {
                        'type': 'prompt',
                        'description': 'Enter role name',
                    }
                },
            },
            '__DONE__': {},
        },
        'execute': {
            'default': [
                {
                    'type': 'apicall',
                    'operation': 'iam.CreateRole',
                    'params': {'RoleName': "{role_name}"},
                    'output_var': 'role_arn',
                    'query': 'Role.Arn',
                }
            ]
        },
    }


@pytest.fixture
def file_prompt_definition():
    return {
        'title': 'Uses API call in prompting stage',
        'plan': {
            'section': {
                'shortname': 'Section',
                'values': {
                    'choose_file': {
                        'description': 'Choose file',
                        'type': 'prompt',
                        'completer': 'file_completer',
                    },
                    'second_prompt': {
                        'description': 'Second prompt',
                        'type': 'prompt',
                    },
                },
            },
            '__DONE__': {},
        },
        'execute': {},
    }


class TestWizardAppRunner(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock(Session)
        self.app_factory = mock.Mock(create_wizard_app)
        self.app = mock.Mock(Application)
        self.app.traverser = mock.Mock(WizardTraverser)
        self.app_factory.return_value = self.app
        self.definition = {}
        self.runner = WizardAppRunner(
            session=self.session, app_factory=self.app_factory
        )

    def test_run_calls_expected_app_interfaces(self):
        self.runner.run(self.definition)
        self.app.run.assert_called_with()
        self.app.traverser.get_output.assert_called_with()

    def test_run_propagates_error_from_app_run(self):
        expected_exception = KeyboardInterrupt
        self.app.run.side_effect = KeyboardInterrupt
        with self.assertRaises(expected_exception):
            self.runner.run(self.definition)


class BaseWizardApplicationTest:
    def assert_app_values(self, app, **expected_values):
        assert dict(app.values) == expected_values

    def assert_buffer_text(self, app, buffer_name, expected_text):
        buffer = app.layout.get_buffer_by_name(buffer_name)
        assert buffer.document.text == expected_text

    def assert_current_buffer(self, app, buffer_name):
        assert app.layout.current_buffer.name, buffer_name

    def assert_expected_buffer_completions(
        self, app, buffer_name, expected_completions
    ):
        buffer = app.layout.get_buffer_by_name(buffer_name)
        assert buffer.complete_state.completions == expected_completions

    def assert_prompt_is_visible(self, app, prompt_name):
        assert prompt_name in self.get_visible_buffers(app)

    def assert_prompt_is_not_visible(self, app, prompt_name):
        assert prompt_name not in self.get_visible_buffers(app)

    def assert_run_wizard_dialog_is_visible(self, app):
        dialog_container = app.layout.run_wizard_dialog.container
        assert dialog_container in self.get_visible_containers(app)

    def assert_run_wizard_dialog_is_not_visible(self, app):
        dialog_container = app.layout.run_wizard_dialog.container
        assert dialog_container not in self.get_visible_containers(app)

    def assert_toolbar_has_text(self, app, text):
        assert any(
            [
                text in tip[1]
                for tip in list(
                    filter(
                        lambda x: getattr(x, 'name', '') == 'toolbar_panel',
                        app.layout.find_all_controls(),
                    )
                )[0].text()()
            ]
        )

    def assert_toolbar_does_not_have_text(self, app, text):
        assert not any(
            [
                text in tip[1]
                for tip in list(
                    filter(
                        lambda x: getattr(x, 'name', '') == 'toolbar_panel',
                        app.layout.find_all_controls(),
                    )
                )[0].text()()
            ]
        )

    def get_visible_buffers(self, app):
        return [
            window.content.buffer.name
            for window in app.layout.visible_windows
            if hasattr(window.content, 'buffer')
        ]

    def get_visible_containers(self, app):
        return list(walk(app.layout.container, skip_hidden=True))


class TestBasicWizardApplication(BaseWizardApplicationTest):
    def test_can_answer_single_prompt(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('val1\n')
            self.assert_app_values(app_runner.app, prompt1='val1')

    def test_can_answer_multiple_prompts(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('val1\n')
            # This is to remove the default answer of foo for prompt 2
            app_runner.feed_input(
                Keys.Backspace, Keys.Backspace, Keys.Backspace
            )
            app_runner.feed_input('val2\n')
            self.assert_app_values(
                app_runner.app, prompt1='val1', prompt2='val2'
            )

    def test_can_use_tab_to_submit_answer(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('val1')
            app_runner.feed_input(Keys.Tab)
            self.assert_app_values(app_runner.app, prompt1='val1')

    def test_can_use_prompt_default_value(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Tab)
            app_runner.feed_input(Keys.Tab)
            self.assert_app_values(app_runner.app, prompt1='', prompt2='foo')

    def test_can_use_shift_tab_to_toggle_backwards_and_change_answer(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('1\n')
            app_runner.feed_input(Keys.BackTab)
            app_runner.feed_input(Keys.Backspace)
            app_runner.feed_input('override1\n')
            self.assert_app_values(
                app_runner.app, prompt1='override1', prompt2='foo'
            )

    def test_can_answer_prompts_across_sections(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('val1\n')
            app_runner.feed_input('\n')
            app_runner.feed_input('second_section_val1\n')
            self.assert_app_values(
                app_runner.app,
                prompt1='val1',
                prompt2='foo',
                second_section_prompt='second_section_val1',
            )

    def test_can_move_back_to_a_prompt_in_previous_section(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('val1\n')
            app_runner.feed_input('\n')
            app_runner.feed_input('second_section_val1\n')
            # At DONE step, hit back tab to select "Go back" button
            # and enter to select that button.
            app_runner.feed_input(Keys.BackTab, Keys.Enter)
            app_runner.feed_input(Keys.BackTab)
            # This is to remove the default answer of foo for prompt 2
            app_runner.feed_input(
                Keys.Backspace, Keys.Backspace, Keys.Backspace
            )
            app_runner.feed_input('override2\n')
            self.assert_app_values(
                app_runner.app,
                prompt1='val1',
                prompt2='override2',
                second_section_prompt='second_section_val1',
            )

    def test_prompts_are_not_visible_across_sections(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            self.assert_prompt_is_visible(app_runner.app, 'prompt1')
            self.assert_prompt_is_visible(app_runner.app, 'prompt2')
            self.assert_prompt_is_not_visible(
                app_runner.app, 'second_section_prompt'
            )

            app_runner.feed_input('val1\n')
            app_runner.feed_input('val2\n')

            self.assert_prompt_is_not_visible(app_runner.app, 'prompt1')
            self.assert_prompt_is_not_visible(app_runner.app, 'prompt2')
            self.assert_prompt_is_visible(
                app_runner.app, 'second_section_prompt'
            )

    def test_run_wizard_dialog_appears_only_at_end_of_wizard(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)

        with app_runner.run_app_in_thread():
            self.assert_run_wizard_dialog_is_not_visible(app_runner.app)
            app_runner.feed_input('val1\n')
            self.assert_run_wizard_dialog_is_not_visible(app_runner.app)
            app_runner.feed_input('val2\n')
            self.assert_run_wizard_dialog_is_not_visible(app_runner.app)
            app_runner.feed_input('second_section_val1\n')
            self.assert_run_wizard_dialog_is_visible(app_runner.app)

    def test_enter_at_wizard_dialog_exits_app_with_zero_rc(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread() as ctx:
            app_runner.feed_input('val1\n')
            app_runner.feed_input('val2\n')
            app_runner.feed_input('second_section_val1\n')
            self.assert_run_wizard_dialog_is_visible(app_runner.app)
            app_runner.feed_input(Keys.Enter)
        assert ctx.return_value == 0

    def test_can_go_back_from_run_wizard_dialog(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('val1\n')
            app_runner.feed_input('\n')
            app_runner.feed_input('second_section_val1\n')
            self.assert_run_wizard_dialog_is_visible(app_runner.app)
            app_runner.feed_input(Keys.Right, Keys.Enter)
            app_runner.feed_input('_override\n')
            self.assert_app_values(
                app_runner.app,
                prompt1='val1',
                prompt2='foo',
                second_section_prompt='second_section_val1_override',
            )

    def test_traverser_get_output(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        assert app_runner.app.traverser.get_output() == 'output: some text'

    def test_can_exit_and_propagate_ctrl_c_from_wizard(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread() as ctx:
            app_runner.feed_input(Keys.ControlC)
        assert isinstance(ctx.raised_exception, KeyboardInterrupt)

    def test_captures_unexpected_errors_when_processing_input(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        unexpected_error = ValueError('Not expected')

        @app_runner.app.key_bindings.add('a')
        def trigger_unexpected_error(event):
            raise unexpected_error

        with app_runner.run_app_in_thread() as ctx:
            app_runner.feed_input('a')
        assert isinstance(ctx.raised_exception, UnexpectedWizardException)
        assert ctx.raised_exception.original_exception is unexpected_error

    def test_cant_open_save_panel_if_no_details_available(
        self, make_stubbed_wizard_runner, basic_definition
    ):
        app_runner = make_stubbed_wizard_runner(basic_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.ControlS)
            assert 'save_details_dialogue' not in self.get_visible_buffers(
                app_runner.app
            )


class TestConditionalWizardApplication(BaseWizardApplicationTest):
    def test_conditional_prompt_is_skipped_when_condition_not_met(
        self, make_stubbed_wizard_runner, conditional_definition
    ):
        app_runner = make_stubbed_wizard_runner(conditional_definition)
        with app_runner.run_app_in_thread():
            self.assert_prompt_is_not_visible(app_runner.app, 'conditional')
            app_runner.feed_input('condition-not-met\n')
            self.assert_prompt_is_not_visible(app_runner.app, 'conditional')
            app_runner.feed_input('after-conditional-val\n')
            self.assert_app_values(
                app_runner.app,
                before_conditional='condition-not-met',
                after_conditional='after-conditional-val',
            )

    def test_conditional_prompt_appears_when_condition_met(
        self, make_stubbed_wizard_runner, conditional_definition
    ):
        app_runner = make_stubbed_wizard_runner(conditional_definition)
        with app_runner.run_app_in_thread():
            self.assert_prompt_is_not_visible(app_runner.app, 'conditional')
            app_runner.feed_input('condition-met\n')
            self.assert_prompt_is_visible(app_runner.app, 'conditional')
            app_runner.feed_input('val-at-conditional\n')
            self.assert_app_values(
                app_runner.app,
                before_conditional='condition-met',
                conditional='val-at-conditional',
            )


class TestChoicesWizardApplication(BaseWizardApplicationTest):
    def test_immediately_pressing_enter_selects_first_choice(
        self, make_stubbed_wizard_runner, choices_definition
    ):
        app_runner = make_stubbed_wizard_runner(choices_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Enter)
            self.assert_app_values(
                app_runner.app, choices_prompt='actual_option_1'
            )

    def test_can_select_choice_in_prompt(
        self, make_stubbed_wizard_runner, choices_definition
    ):
        app_runner = make_stubbed_wizard_runner(choices_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Down, Keys.Enter)
            self.assert_app_values(
                app_runner.app, choices_prompt='actual_option_2'
            )


class TestCorruptedChoicesWizardApplication(BaseWizardApplicationTest):
    def test_can_handle_corrupted_choices(
        self, make_stubbed_wizard_runner, corrupted_choice_definition
    ):
        app_runner = make_stubbed_wizard_runner(corrupted_choice_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Down, Keys.Enter)
            self.assert_app_values(app_runner.app, choices_prompt=None)

    def test_show_error_message_on_get_value_exception(
        self, make_stubbed_wizard_runner, corrupted_choice_definition
    ):
        app_runner = make_stubbed_wizard_runner(corrupted_choice_definition)
        with app_runner.run_app_in_thread():
            self.assert_buffer_text(
                app_runner.app,
                'error_bar',
                'Encountered following error in wizard:\n\n\'operation\'',
            )

    def test_toolbar_text_on_get_value_exception(
        self, make_stubbed_wizard_runner, corrupted_choice_definition
    ):
        app_runner = make_stubbed_wizard_runner(corrupted_choice_definition)
        with app_runner.run_app_in_thread():
            self.assert_toolbar_has_text(app_runner.app, 'error message')


class TestMixedPromptTypeWizardApplication(BaseWizardApplicationTest):
    def test_can_answer_buffer_prompt_followed_by_select_prompt(
        self, make_stubbed_wizard_runner, mixed_prompt_definition
    ):
        app_runner = make_stubbed_wizard_runner(mixed_prompt_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('buffer_answer\n')
            app_runner.feed_input(Keys.Enter)
            self.assert_app_values(
                app_runner.app,
                buffer_input_prompt='buffer_answer',
                select_prompt='select_answer_1',
            )


class TestPromptWithDataConvertWizardApplication(BaseWizardApplicationTest):
    def test_can_convert_integers(
        self, make_stubbed_wizard_runner, data_convert_definition
    ):
        app_runner = make_stubbed_wizard_runner(data_convert_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('100\n')
            self.assert_app_values(app_runner.app, buffer_input_int=100)

    def test_can_convert_bool(
        self, make_stubbed_wizard_runner, data_convert_definition
    ):
        app_runner = make_stubbed_wizard_runner(data_convert_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('0\n')
            app_runner.feed_input('true\n')
            self.assert_app_values(
                app_runner.app, buffer_input_int=0, buffer_input_bool=True
            )

    def test_show_error_and_stop_on_incorrect_input(
        self, make_stubbed_wizard_runner, data_convert_definition
    ):
        app_runner = make_stubbed_wizard_runner(data_convert_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('foo\n')
            self.assert_buffer_text(
                app_runner.app,
                'error_bar',
                'Encountered following error in wizard:\n\n'
                'Invalid value foo for datatype int',
            )
            self.assert_current_buffer(app_runner.app, 'buffer_input_int')

    def test_clear_error_and_go_on_after_correction(
        self, make_stubbed_wizard_runner, data_convert_definition
    ):
        app_runner = make_stubbed_wizard_runner(data_convert_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('foo\n')
            self.assert_buffer_text(
                app_runner.app,
                'error_bar',
                'Encountered following error in wizard:\n\n'
                'Invalid value foo for datatype int',
            )
            self.assert_current_buffer(app_runner.app, 'buffer_input_int')
            app_runner.feed_input(
                Keys.Backspace, Keys.Backspace, Keys.Backspace
            )
            app_runner.feed_input('100\n')
            self.assert_buffer_text(app_runner.app, 'error_bar', '')
            self.assert_current_buffer(app_runner.app, 'buffer_input_bool')


class TestApiCallWizardApplication(BaseWizardApplicationTest):
    def test_uses_choices_from_api_call(
        self, make_stubbed_wizard_runner, api_call_definition, mock_iam_client
    ):
        app_runner = make_stubbed_wizard_runner(api_call_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Enter)
            self.assert_app_values(app_runner.app, choose_policy='policy1_arn')


class TestDetailsWizardApplication(BaseWizardApplicationTest):
    def test_get_details_for_choice(
        self, make_stubbed_wizard_runner, details_definition, mock_iam_client
    ):
        app_runner = make_stubbed_wizard_runner(details_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.F3)
            self.assert_buffer_text(
                app_runner.app,
                'details_buffer',
                '{\n    "policy": "policy_document"\n}',
            )

    def test_details_disabled_for_choice_wo_details(
        self, make_stubbed_wizard_runner, details_definition, mock_iam_client
    ):
        app_runner = make_stubbed_wizard_runner(details_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Tab)
            self.assert_prompt_is_not_visible(
                app_runner.app, 'toolbar_details'
            )
            app_runner.feed_input(Keys.F3)
            self.assert_prompt_is_not_visible(app_runner.app, 'details_buffer')

    def test_can_switch_focus_to_details_panel(
        self, make_stubbed_wizard_runner, details_definition, mock_iam_client
    ):
        app_runner = make_stubbed_wizard_runner(details_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.F3)
            app_runner.feed_input(Keys.F2)
            self.assert_current_buffer(app_runner.app, 'details_buffer')
            app_runner.feed_input(Keys.F3)
            assert app_runner.app.layout.current_buffer is None

    def test_can_toggle_save_panel(
        self, make_stubbed_wizard_runner, details_definition, mock_iam_client
    ):
        app_runner = make_stubbed_wizard_runner(details_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.ControlS)
            self.assert_current_buffer(app_runner.app, 'save_details_dialogue')
            self.assert_prompt_is_visible(
                app_runner.app, 'save_details_dialogue'
            )
            assert app_runner.app.details_visible
            assert app_runner.app.save_details_visible

    def test_can_not_switch_focus_to_details_panel_if_it_not_visible(
        self, make_stubbed_wizard_runner, details_definition, mock_iam_client
    ):
        app_runner = make_stubbed_wizard_runner(details_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.F2)
            assert app_runner.app.layout.current_buffer is None

    def test_can_set_details_panel_title(
        self, make_stubbed_wizard_runner, details_definition, mock_iam_client
    ):
        app_runner = make_stubbed_wizard_runner(details_definition)
        with app_runner.run_app_in_thread():
            assert app_runner.app.details_title == 'Policy Document'
            app_runner.feed_input(Keys.F3)
            assert app_runner.app.details_title == 'Policy Document'

    def test_can_set_details_toolbar_text(
        self, make_stubbed_wizard_runner, details_definition, mock_iam_client
    ):
        app_runner = make_stubbed_wizard_runner(details_definition)
        with app_runner.run_app_in_thread():
            self.assert_toolbar_has_text(app_runner.app, 'Policy Document')


class TestPreviewWizardApplication(BaseWizardApplicationTest):
    def test_details_panel_visible_by_default(
        self, make_stubbed_wizard_runner, preview_definition
    ):
        app_runner = make_stubbed_wizard_runner(preview_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Tab)
            self.assert_prompt_is_visible(app_runner.app, 'details_buffer')

    def test_get_details_for_choice(
        self, make_stubbed_wizard_runner, preview_definition
    ):
        app_runner = make_stubbed_wizard_runner(preview_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Tab)
            self.assert_buffer_text(
                app_runner.app, 'details_buffer', 'First option details'
            )

    def test_get_details_for_second_choice(
        self, make_stubbed_wizard_runner, preview_definition
    ):
        app_runner = make_stubbed_wizard_runner(preview_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Tab)
            app_runner.feed_input(Keys.Down)
            self.assert_buffer_text(
                app_runner.app, 'details_buffer', 'Second option details'
            )


class TestSharedConfigWizardApplication(BaseWizardApplicationTest):
    def test_uses_choices_from_api_call(
        self,
        make_stubbed_wizard_runner,
        shared_config_definition,
        mock_botocore_session,
    ):
        mock_botocore_session.available_profiles = ['profile1', 'profile2']
        app_runner = make_stubbed_wizard_runner(shared_config_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.Enter)
            self.assert_app_values(app_runner.app, choose_profile='profile1')


class TestRunWizardApplication(BaseWizardApplicationTest):
    def assert_error_bar_is_visible(self, app):
        assert 'error_bar' in self.get_visible_buffers(app)

    def assert_error_bar_is_not_visible(self, app):
        assert 'error_bar' not in self.get_visible_buffers(app)

    def test_run_wizard_execute(
        self,
        make_stubbed_wizard_runner,
        run_wizard_definition,
        mock_iam_client,
    ):
        app_runner = make_stubbed_wizard_runner(run_wizard_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('role-name\n')
            app_runner.feed_input(Keys.Enter)
        mock_iam_client.create_role.assert_called_with(RoleName='role-name')
        assert app_runner.app.values['role_arn'] == 'returned-role-arn'

    def test_run_wizard_captures_and_displays_errors(
        self,
        make_stubbed_wizard_runner,
        run_wizard_definition,
        mock_iam_client,
    ):
        app_runner = make_stubbed_wizard_runner(run_wizard_definition)
        mock_iam_client.create_role.side_effect = Exception(
            'Error creating role'
        )
        with app_runner.run_app_in_thread():
            app_runner.feed_input('role-name\n')
            app_runner.feed_input(Keys.Enter)
            self.assert_buffer_text(
                app_runner.app,
                'error_bar',
                'Encountered following error in wizard:\n\nError creating role',
            )

    def test_can_change_answers_on_run_wizard_failure(
        self,
        make_stubbed_wizard_runner,
        run_wizard_definition,
        mock_iam_client,
    ):
        app_runner = make_stubbed_wizard_runner(run_wizard_definition)
        mock_iam_client.create_role.side_effect = [
            Exception('Initial error'),
            mock_iam_client.create_role.return_value,
        ]
        with app_runner.run_app_in_thread():
            app_runner.feed_input('role-name\n')
            self.assert_error_bar_is_not_visible(app_runner.app)
            self.assert_run_wizard_dialog_is_visible(app_runner.app)
            app_runner.feed_input(Keys.Enter)
            self.assert_buffer_text(
                app_runner.app,
                'error_bar',
                'Encountered following error in wizard:\n\nInitial error',
            )

            # Make sure dialog closed on error but error message is visible
            self.assert_run_wizard_dialog_is_not_visible(app_runner.app)
            self.assert_error_bar_is_visible(app_runner.app)
            app_runner.feed_input('-new\n')

            # Enter "Yes" to the done dialog again
            app_runner.feed_input(Keys.Enter)

        mock_iam_client.create_role.assert_called_with(
            RoleName='role-name-new'
        )
        assert app_runner.app.values['role_arn'] == 'returned-role-arn'

    def test_can_switch_exception_panel(
        self,
        make_stubbed_wizard_runner,
        run_wizard_definition,
        mock_iam_client,
    ):
        app_runner = make_stubbed_wizard_runner(run_wizard_definition)
        mock_iam_client.create_role.side_effect = [
            Exception('Initial error'),
            mock_iam_client.create_role.return_value,
        ]
        with app_runner.run_app_in_thread():
            app_runner.feed_input('role-name\n')
            self.assert_error_bar_is_not_visible(app_runner.app)
            app_runner.feed_input(Keys.Enter)
            self.assert_error_bar_is_visible(app_runner.app)
            app_runner.feed_input(Keys.F4)
            self.assert_error_bar_is_not_visible(app_runner.app)

    def test_toolbar_has_exception_panel_hot_key(
        self,
        make_stubbed_wizard_runner,
        run_wizard_definition,
        mock_iam_client,
    ):
        app_runner = make_stubbed_wizard_runner(run_wizard_definition)
        mock_iam_client.create_role.side_effect = [
            Exception('Initial error'),
            mock_iam_client.create_role.return_value,
        ]
        with app_runner.run_app_in_thread():
            app_runner.feed_input('role-name\n')
            self.assert_error_bar_is_not_visible(app_runner.app)
            self.assert_toolbar_does_not_have_text(
                app_runner.app, 'error message'
            )
            app_runner.feed_input(Keys.Enter)
            self.assert_error_bar_is_visible(app_runner.app)
            self.assert_toolbar_has_text(app_runner.app, 'error message')


class TestWizardTraverser(unittest.TestCase):
    def setUp(self):
        self.simple_definition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'first_prompt': self.create_prompt_definition(),
                        'second_prompt': self.create_prompt_definition(),
                    }
                }
            }
        )
        self.simple_choice_definition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'choice_prompt': self.create_prompt_definition(
                            choices=[
                                {
                                    'display': 'Option 1',
                                    'actual_value': 'actual_option_1',
                                },
                                {
                                    'display': 'Option 2',
                                    'actual_value': 'actual_option_2',
                                },
                            ]
                        ),
                    }
                }
            }
        )

        self.single_prompt_definition = self.create_definition(
            sections={
                'only_section': {
                    'values': {
                        'only_prompt': self.create_prompt_definition(),
                    }
                }
            }
        )

        self.single_prompt_definition_with_datatype = self.create_definition(
            sections={
                'only_section': {
                    'values': {
                        'only_prompt': self.create_prompt_definition(
                            datatype='int'
                        ),
                    }
                }
            }
        )

    def create_traverser(
        self,
        definition,
        values=None,
        executor=None,
    ):
        if values is None:
            values = {}
        if executor is None:
            executor = mock.Mock(Executor)
        return WizardTraverser(definition, values, executor)

    def create_definition(self, sections, execute=None):
        sections = sections.copy()
        sections['__DONE__'] = {}
        if execute is None:
            execute = {}
        return {'plan': sections, 'execute': execute}

    def create_prompt_definition(
        self, description=None, condition=None, choices=None, datatype=None
    ):
        prompt = {'type': 'prompt'}
        if not description:
            description = 'A sample description'
        prompt['description'] = description
        if condition:
            prompt['condition'] = condition
        if choices:
            prompt['choices'] = choices
        if datatype:
            prompt['datatype'] = datatype
        return prompt

    def test_get_current_prompt(self):
        traverser = self.create_traverser(self.simple_definition)
        self.assertEqual(traverser.get_current_prompt(), 'first_prompt')

    def test_get_current_prompt_returns_done_when_no_more_prompts(self):
        traverser = self.create_traverser(self.single_prompt_definition)
        traverser.next_prompt()
        self.assertEqual(traverser.get_current_prompt(), traverser.DONE)

    def test_get_current_section(self):
        traverser = self.create_traverser(self.simple_definition)
        self.assertEqual(traverser.get_current_section(), 'first_section')

    def test_get_current_section_returns_done_when_no_more_prompts(self):
        traverser = self.create_traverser(self.single_prompt_definition)
        traverser.next_prompt()
        self.assertEqual(traverser.get_current_section(), traverser.DONE)

    def test_get_current_prompt_choices(self):
        traverser = self.create_traverser(self.simple_choice_definition)
        self.assertEqual(
            traverser.get_current_prompt_choices(), ['Option 1', 'Option 2']
        )

    def test_get_current_prompt_choices_as_list_of_strings(self):
        choice_wizard = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'choice_prompt': self.create_prompt_definition(
                            choices=[
                                'from_list_1',
                                'from_list_2',
                            ]
                        ),
                    }
                }
            }
        )
        traverser = self.create_traverser(choice_wizard)
        self.assertEqual(
            traverser.get_current_prompt_choices(),
            ['from_list_1', 'from_list_2'],
        )

    def test_get_current_prompt_choices_from_variable(self):
        values = {'choices_var': ['from_var_1', 'from_var_2']}
        choice_wizard = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'choice_prompt': self.create_prompt_definition(
                            choices='choices_var'
                        ),
                    }
                }
            }
        )
        traverser = self.create_traverser(choice_wizard, values)
        self.assertEqual(
            traverser.get_current_prompt_choices(),
            ['from_var_1', 'from_var_2'],
        )

    def test_get_current_prompt_choices_returns_none_when_no_choices(self):
        traverser = self.create_traverser(self.simple_definition)
        self.assertEqual(traverser.get_current_prompt_choices(), None)

    def test_submit_prompt_answer(self):
        values = {}
        traverser = self.create_traverser(self.simple_definition, values)
        traverser.submit_prompt_answer('answer1')
        self.assertEqual(values, {'first_prompt': 'answer1'})

    def test_submit_prompt_answer_converts_display_value(self):
        values = {}
        traverser = self.create_traverser(
            self.simple_choice_definition, values
        )
        traverser.submit_prompt_answer('Option 1')
        self.assertEqual(values, {'choice_prompt': 'actual_option_1'})

    def test_submit_prompt_answer_converts_datatype(self):
        values = {}
        traverser = self.create_traverser(
            self.single_prompt_definition_with_datatype, values
        )
        traverser.submit_prompt_answer('100')
        self.assertEqual(values, {'only_prompt': 100})

    def test_submit_prompt_raise_correct_exception_on_datatype_convert(self):
        values = {}
        traverser = self.create_traverser(
            self.single_prompt_definition_with_datatype,
            values,
        )
        with self.assertRaises(InvalidDataTypeConversionException) as e:
            traverser.submit_prompt_answer('foo')
        self.assertEqual(
            str(e.exception), 'Invalid value foo for datatype int'
        )

    def test_submit_prompt_answer_throw_error_for_invalid_option(self):
        traverser = self.create_traverser(self.simple_choice_definition)
        with self.assertRaises(InvalidChoiceException):
            traverser.submit_prompt_answer('Not an option')

    def test_next_prompt(self):
        traverser = self.create_traverser(self.simple_definition)
        next_prompt = traverser.next_prompt()
        self.assertEqual(next_prompt, 'second_prompt')
        self.assertEqual(traverser.get_current_section(), 'first_section')

    def test_next_prompt_condition_not_met(self):
        definition_with_condition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'first_prompt': self.create_prompt_definition(),
                        'conditional': self.create_prompt_definition(
                            condition={
                                'variable': 'first_prompt',
                                'equals': 'condition-met',
                            }
                        ),
                        'after_conditional': self.create_prompt_definition(),
                    }
                }
            }
        )
        traverser = self.create_traverser(
            definition_with_condition,
            values={'first_prompt': 'condition-not-met'},
        )
        next_prompt = traverser.next_prompt()
        self.assertEqual(next_prompt, 'after_conditional')

    def test_next_prompt_condition_met(self):
        definition_with_condition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'first_prompt': self.create_prompt_definition(),
                        'conditional': self.create_prompt_definition(
                            condition={
                                'variable': 'first_prompt',
                                'equals': 'condition-met',
                            }
                        ),
                        'after_conditional': self.create_prompt_definition(),
                    }
                }
            }
        )
        traverser = self.create_traverser(
            definition_with_condition, values={'first_prompt': 'condition-met'}
        )
        next_prompt = traverser.next_prompt()
        self.assertEqual(next_prompt, 'conditional')

    def test_next_prompt_moves_to_next_section_when_no_more_prompts(self):
        definition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'first_prompt': self.create_prompt_definition(),
                    }
                },
                'second_section': {
                    'values': {
                        'second_prompt': self.create_prompt_definition(),
                    }
                },
            }
        )
        traverser = self.create_traverser(definition)
        next_prompt = traverser.next_prompt()
        self.assertEqual(next_prompt, 'second_prompt')
        self.assertEqual(traverser.get_current_section(), 'second_section')

    def test_next_prompt_returns_done_at_end_of_wizard(self):
        traverser = self.create_traverser(self.single_prompt_definition)
        next_prompt = traverser.next_prompt()
        self.assertEqual(next_prompt, traverser.DONE)

    def test_previous_prompt(self):
        traverser = self.create_traverser(self.simple_definition)
        next_prompt = traverser.next_prompt()
        self.assertEqual(next_prompt, 'second_prompt')
        self.assertEqual(traverser.get_current_section(), 'first_section')

        previous_prompt = traverser.previous_prompt()
        self.assertEqual(previous_prompt, 'first_prompt')
        self.assertEqual(traverser.get_current_section(), 'first_section')

    def test_previous_prompt_can_move_back_sections(self):
        definition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'first_prompt': self.create_prompt_definition(),
                    }
                },
                'second_section': {
                    'values': {
                        'second_prompt': self.create_prompt_definition(),
                    }
                },
            }
        )
        traverser = self.create_traverser(definition)
        next_prompt = traverser.next_prompt()
        self.assertEqual(next_prompt, 'second_prompt')
        self.assertEqual(traverser.get_current_section(), 'second_section')

        previous_prompt = traverser.previous_prompt()
        self.assertEqual(previous_prompt, 'first_prompt')
        self.assertEqual(traverser.get_current_section(), 'first_section')

    def test_previous_prompt_returns_same_prompt_at_start_of_wizard(self):
        traverser = self.create_traverser(self.simple_definition)
        previous_prompt = traverser.previous_prompt()
        self.assertEqual(previous_prompt, 'first_prompt')
        self.assertEqual(traverser.get_current_section(), 'first_section')

    def test_previous_prompt_returns_last_prompt_from_done(self):
        traverser = self.create_traverser(self.single_prompt_definition)
        traverser.next_prompt()
        self.assertTrue(traverser.has_no_remaining_prompts())
        self.assertEqual(traverser.previous_prompt(), 'only_prompt')

    def test_run_wizard_calls_executor(self):
        executor = mock.Mock(Executor)
        values = {}
        execute_definition = {'step': []}
        traverser = self.create_traverser(
            definition=self.create_definition(
                sections={
                    'only_section': {
                        'values': {
                            'only_prompt': self.create_prompt_definition(),
                        }
                    },
                },
                execute=execute_definition,
            ),
            values=values,
            executor=executor,
        )
        traverser.submit_prompt_answer('only_answer')
        traverser.next_prompt()
        traverser.run_wizard()
        executor.execute.assert_called_with(execute_definition, values)

    def test_run_wizard_throws_error_when_not_at_done_step(self):
        traverser = self.create_traverser(self.simple_definition)
        with self.assertRaises(UnableToRunWizardError):
            traverser.run_wizard()

    def test_prompt_is_visible_when_no_condition(self):
        traverser = self.create_traverser(self.simple_definition)
        self.assertTrue(traverser.is_prompt_visible('second_prompt'))

    def test_prompt_is_visible_when_condition_met(self):
        definition_with_condition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'first_prompt': self.create_prompt_definition(),
                        'conditional': self.create_prompt_definition(
                            condition={
                                'variable': 'first_prompt',
                                'equals': 'condition-met',
                            }
                        ),
                    }
                }
            }
        )
        traverser = self.create_traverser(
            definition_with_condition,
            values={'first_prompt': 'condition-not-met'},
        )
        self.assertFalse(traverser.is_prompt_visible('conditional'))

    def test_prompt_is_not_visible_when_condition_not_met(self):
        definition_with_condition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'first_prompt': self.create_prompt_definition(),
                        'conditional': self.create_prompt_definition(
                            condition={
                                'variable': 'first_prompt',
                                'equals': 'condition-met',
                            }
                        ),
                    }
                }
            }
        )
        traverser = self.create_traverser(
            definition_with_condition, values={'first_prompt': 'condition-met'}
        )
        self.assertTrue(traverser.is_prompt_visible('conditional'))

    def test_prompt_is_not_visible_when_not_on_current_section(self):
        definition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'first_prompt': self.create_prompt_definition(),
                    }
                },
                'second_section': {
                    'values': {
                        'second_prompt': self.create_prompt_definition(),
                    }
                },
            }
        )
        traverser = self.create_traverser(definition)
        self.assertFalse(traverser.is_prompt_visible('second_prompt'))

    def test_has_visited_section(self):
        definition = self.create_definition(
            sections={
                'first_section': {
                    'values': {
                        'first_prompt': self.create_prompt_definition(),
                    }
                },
                'second_section': {
                    'values': {
                        'second_prompt': self.create_prompt_definition(),
                    }
                },
                'third_section': {
                    'values': {
                        'third_prompt': self.create_prompt_definition(),
                    }
                },
            }
        )
        # Start at the first section
        traverser = self.create_traverser(definition)
        self.assertTrue(traverser.has_visited_section('first_section'))
        self.assertFalse(traverser.has_visited_section('second_section'))
        self.assertFalse(traverser.has_visited_section('third_section'))

        # Move to the second section
        traverser.next_prompt()
        self.assertTrue(traverser.has_visited_section('first_section'))
        self.assertTrue(traverser.has_visited_section('second_section'))
        self.assertFalse(traverser.has_visited_section('third_section'))

        # Go back to the first section. Note the second section should still
        # count as visited.
        traverser.previous_prompt()
        self.assertTrue(traverser.has_visited_section('first_section'))
        self.assertTrue(traverser.has_visited_section('second_section'))
        self.assertFalse(traverser.has_visited_section('third_section'))

    def test_has_no_remaining_prompts(self):
        traverser = self.create_traverser(self.single_prompt_definition)
        self.assertFalse(traverser.has_no_remaining_prompts())
        traverser.next_prompt()
        self.assertTrue(traverser.has_no_remaining_prompts())


class TestWizardValues(unittest.TestCase):
    def setUp(self):
        self.definition = {
            'title': 'Wizard',
            'plan': {
                'section': {
                    'shortname': 'Section',
                    'values': {
                        'no_handler_value': {'type': 'prompt'},
                        'handler_value': {'type': 'use_handler'},
                    },
                }
            },
        }
        self.handler = mock.Mock(BaseStep)
        self.exception_handler = mock.Mock()

    def get_wizard_values(self, retrieval_steps=None, exception_handler=None):
        if exception_handler is None:
            exception_handler = self.exception_handler
        return WizardValues(
            self.definition, retrieval_steps, exception_handler
        )

    def test_manual_get_and_set(self):
        values = self.get_wizard_values()
        values['no_handler_value'] = 'manual'
        self.assertEqual(values['no_handler_value'], 'manual')

    def test_get_with_handler(self):
        self.handler.run_step.return_value = 'from_handler'
        values = self.get_wizard_values({'use_handler': self.handler})
        self.assertEqual(values['handler_value'], 'from_handler')

    def test_manual_set_overrides_delegation_handler(self):
        self.handler.run_step.return_value = 'from_handler'
        values = self.get_wizard_values({'use_handler': self.handler})
        values['handler_value'] = 'manual'
        self.assertEqual(values['handler_value'], 'manual')
        self.handler.run_step.assert_not_called()

    def test_delete(self):
        values = self.get_wizard_values()
        values['no_handler_value'] = 'manual'
        self.assertEqual(values['no_handler_value'], 'manual')
        del values['no_handler_value']
        self.assertNotIn('no_handler_value', values)

    def test_iterates_over_value_names(self):
        values = self.get_wizard_values()
        values['no_handler_value'] = 'manual'
        value_names = list(values)
        self.assertEqual(value_names, ['no_handler_value'])

    def test_length(self):
        values = self.get_wizard_values()
        values['no_handler_value'] = 'manual'
        self.assertEqual(len(values), 1)

    def test_copy(self):
        values = self.get_wizard_values()
        values['foo'] = 'bar'
        values['bar'] = 'foo'
        values_copy = values.copy()
        for key in values.keys():
            self.assertEqual(values[key], values_copy[key])
        self.assertEqual(len(values), len(values_copy))
        self.assertNotEqual(id(values), id(values_copy))

    def test_exception_handler(self):
        e = Exception()
        self.handler.run_step.side_effect = e
        values = self.get_wizard_values({'use_handler': self.handler})
        value = values['handler_value']
        self.exception_handler.assert_called_with(e)


class TestPromptCompletionWizardApplication(BaseWizardApplicationTest):
    def test_show_completions_for_buffer_text(
        self,
        make_stubbed_wizard_runner,
        file_prompt_definition,
        patch_path_completer,
    ):
        app_runner = make_stubbed_wizard_runner(file_prompt_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('fi')
            self.assert_expected_buffer_completions(
                app_runner.app,
                'choose_file',
                [
                    Completion('le1', 0, display='file1'),
                    Completion('le2', 0, display='file2'),
                ],
            )

    def test_choose_completion_on_Enter_and_stays_on_the_same_prompt(
        self,
        make_stubbed_wizard_runner,
        file_prompt_definition,
        patch_path_completer,
    ):
        app_runner = make_stubbed_wizard_runner(file_prompt_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('fi')
            app_runner.feed_input(Keys.Down, Keys.Enter)
            self.assert_buffer_text(app_runner.app, 'choose_file', 'file1')
            self.assert_current_buffer(app_runner.app, 'choose_file')

    def test_choose_completion_on_Enter_and_move_on_second_Enter(
        self,
        make_stubbed_wizard_runner,
        file_prompt_definition,
        patch_path_completer,
    ):
        app_runner = make_stubbed_wizard_runner(file_prompt_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('fi')
            app_runner.feed_input(Keys.Down, Keys.Enter)
            app_runner.feed_input(Keys.Enter)
            self.assert_current_buffer(app_runner.app, 'second_prompt')

    def test_switch_completions_on_tab(
        self,
        make_stubbed_wizard_runner,
        file_prompt_definition,
        patch_path_completer,
    ):
        app_runner = make_stubbed_wizard_runner(file_prompt_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('fi')
            app_runner.feed_input(Keys.Tab, Keys.Tab, Keys.Enter)
            self.assert_buffer_text(app_runner.app, 'choose_file', 'file2')
            self.assert_current_buffer(app_runner.app, 'choose_file')

    def test_switch_completions_on_back_tab(
        self,
        make_stubbed_wizard_runner,
        file_prompt_definition,
        patch_path_completer,
    ):
        app_runner = make_stubbed_wizard_runner(file_prompt_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('fi')
            app_runner.feed_input(Keys.Tab, Keys.Tab, Keys.BackTab, Keys.Enter)
            self.assert_buffer_text(app_runner.app, 'choose_file', 'file1')
            self.assert_current_buffer(app_runner.app, 'choose_file')

    def test_switch_prompt_on_tab_if_it_is_not_completing(
        self,
        make_stubbed_wizard_runner,
        file_prompt_definition,
        patch_path_completer,
    ):
        app_runner = make_stubbed_wizard_runner(file_prompt_definition)
        with app_runner.run_app_in_thread():
            app_runner.feed_input('rrrr')
            app_runner.feed_input(Keys.Tab)
            self.assert_current_buffer(app_runner.app, 'second_prompt')


class TestSaveDetailsWizard(BaseWizardApplicationTest):
    def test_file_saved_to_disk(self, make_stubbed_wizard_runner):
        definition = {
            'title': 'Show details in prompting stage',
            'plan': {
                'section': {
                    'shortname': 'Section',
                    'values': {
                        'policy_arn': {
                            'description': 'Choose policy',
                            'type': 'prompt',
                            'choices': ['a'],
                            'details': {
                                'value': 'test_value',
                                'description': 'Policy Document',
                                'output': 'json',
                            },
                        },
                    },
                },
                '__DONE__': {},
            },
            'execute': {},
        }
        app_runner = make_stubbed_wizard_runner(definition)
        mock_file_io = mock.Mock(spec=FileIO)
        app_runner.app.file_io = mock_file_io
        app_runner.app.values['test_value'] = {'foo': 'bar'}
        details_content = '{\n    "foo": "bar"\n}'
        with app_runner.run_app_in_thread():
            app_runner.feed_input(Keys.F3)
            self.assert_buffer_text(
                app_runner.app, 'details_buffer', details_content
            )
            app_runner.feed_input(Keys.ControlS)
            app_runner.feed_input('/tmp/myfile.json\n')
            mock_file_io.write_file_contents.assert_called_with(
                '/tmp/myfile.json', details_content
            )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestPutConfigurationRecorder(BaseAWSCommandParamsTest):
    prefix = 'configservice put-configuration-recorder'

    def test_only_configuration_recorder(self):
        cmdline = self.prefix + ' --configuration-recorder'
        cmdline += ' name=myrecorder,roleARN=myarn'
        result = {
            'ConfigurationRecorder': {'name': 'myrecorder', 'roleARN': 'myarn'}
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_no_configuration_recorder(self):
        stdout, stderr, rc = self.run_cmd(self.prefix, expected_rc=252)
        self.assertIn('required', stderr)
        self.assertIn('--configuration-recorder', stderr)

    def test_configuration_recorder_with_recording_group(self):
        cmdline = self.prefix + ' --configuration-recorder'
        cmdline += ' name=myrecorder,roleARN=myarn'
        cmdline += ' --recording-group'
        cmdline += ' allSupported=true,resourceTypes='
        cmdline += 'AWS::EC2::Volume,AWS::EC2::VPC'
        result = {
            'ConfigurationRecorder': {
                'name': 'myrecorder',
                'roleARN': 'myarn',
                'recordingGroup': {
                    'allSupported': True,
                    'resourceTypes': ['AWS::EC2::Volume', 'AWS::EC2::VPC'],
                },
            }
        }
        self.assert_params_for_cmd(cmdline, result)



#!/usr/bin/env python
# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime

from ruamel.yaml import YAML

from awscli.compat import ensure_text_type
from awscli.testutils import BaseAWSCommandParamsTest, skip_if_windows


class TestYAMLOutput(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestYAMLOutput, self).setUp()
        self.parsed_response = {
            'Users': [
                {
                    "UserName": "testuser-50",
                    "Path": "/",
                    "CreateDate": "2013-02-12T19:08:52Z",
                    "UserId": "EXAMPLEUSERID",
                    "Arn": "arn:aws:iam::12345:user/testuser1",
                },
                {
                    "UserName": "testuser-51",
                    "Path": "/",
                    "CreateDate": "2012-10-14T23:53:39Z",
                    "UserId": "EXAMPLEUSERID",
                    "Arn": "arn:aws:iam::123456:user/testuser2",
                },
            ]
        }

        self.yaml = YAML(typ="safe", pure=True)

    def test_yaml_response(self):
        stdout, _, _ = self.run_cmd(
            'iam list-users --output yaml', expected_rc=0
        )
        expected = (
            "Users:\n"
            "- Arn: arn:aws:iam::12345:user/testuser1\n"
            "  CreateDate: '2013-02-12T19:08:52Z'\n"
            "  Path: /\n"
            "  UserId: EXAMPLEUSERID\n"
            "  UserName: testuser-50\n"
            "- Arn: arn:aws:iam::123456:user/testuser2\n"
            "  CreateDate: '2012-10-14T23:53:39Z'\n"
            "  Path: /\n"
            "  UserId: EXAMPLEUSERID\n"
            "  UserName: testuser-51\n"
        )
        self.assertEqual(stdout, expected)
        parsed_output = self.yaml.load(stdout)
        self.assertEqual(self.parsed_response, parsed_output)

    def test_empty_dict_response_prints_nothing(self):
        self.parsed_response = {}
        stdout, _, _ = self.run_cmd(
            'sts get-caller-identity --output yaml', expected_rc=0
        )
        self.assertEqual(stdout, '')

    def test_empty_list_prints_list(self):
        self.parsed_response = []
        stdout, _, _ = self.run_cmd(
            'sts get-caller-identity --output yaml', expected_rc=0
        )
        self.assertEqual(stdout, '[]\n')

    def test_empty_string_prints_nothing(self):
        self.parsed_response = ''
        stdout, _, _ = self.run_cmd(
            'sts get-caller-identity --output yaml', expected_rc=0
        )
        self.assertEqual(stdout, '""\n')

    def test_jmespath_yaml_response(self):
        jmespath_query = 'Users[*].UserName'
        stdout, _, _ = self.run_cmd(
            'iam list-users --output yaml --query %s' % jmespath_query,
            expected_rc=0,
        )
        parsed_output = self.yaml.load(stdout)
        self.assertEqual(parsed_output, ['testuser-50', 'testuser-51'])

    def test_print_int(self):
        self.parsed_response = {'Count': 1}
        command = (
            'dynamodb scan --output yaml --table-name foo --select COUNT '
            '--query Count'
        )
        stdout, _, _ = self.run_cmd(command, expected_rc=0)
        self.assertEqual(stdout, '1\n')

    def test_print_float(self):
        self.parsed_response = {'ConsumedCapacity': {'CapacityUnits': 0.5}}
        command = (
            'dynamodb scan --output yaml --table-name foo '
            '--query ConsumedCapacity.CapacityUnits'
        )
        stdout, _, _ = self.run_cmd(command, expected_rc=0)
        self.assertEqual(stdout, '0.5\n')

    def test_print_none(self):
        self.parsed_response = {'ConsumedCapacity': None}
        command = (
            'dynamodb scan --output yaml --table-name foo --query '
            'ConsumedCapacity'
        )
        stdout, _, _ = self.run_cmd(command, expected_rc=0)
        self.assertEqual(stdout, 'null\n')

    def test_print_bool(self):
        self.parsed_response = {
            "Contents": [],
            "IsTruncated": False,
        }
        command = (
            's3api list-objects --output yaml --no-paginate '
            '--bucket foo --query IsTruncated'
        )
        stdout, _, _ = self.run_cmd(command, expected_rc=0)
        self.assertEqual(stdout, 'false\n')

    def test_print_timestamp(self):
        now = datetime.datetime(1970, 1, 1, 0, 0)
        self.parsed_response = {
            "Buckets": [
                {
                    "CreationDate": now,
                    "Name": "foo",
                }
            ]
        }
        command = (
            's3api list-buckets --output yaml --query Buckets[0].CreationDate'
        )
        stdout, _, _ = self.run_cmd(command)
        expected = '1970-01-01T00:00:00'
        self.assertIn(expected, stdout)

    def test_print_string_literal(self):
        jmespath_query = 'Users[0].UserName'
        stdout, _, _ = self.run_cmd(
            'iam list-users --output yaml --query %s' % jmespath_query,
            expected_rc=0,
        )
        self.assertEqual(stdout, '"testuser-50"\n')

    @skip_if_windows('Encoding tests only supported on mac/linux')
    def test_yaml_prints_unicode_chars(self):
        self.parsed_response['Users'][1]['UserId'] = '\u2713'
        stdout, _, _ = self.run_cmd(
            'iam list-users --output yaml', expected_rc=0
        )
        self.assertIn('\u2713', ensure_text_type(stdout))

    @skip_if_windows('Encoding tests only supported on mac/linux')
    def test_unicode_yaml_scalar(self):
        self.parsed_response['Users'][1]['UserId'] = '\u2713'
        stdout, _, _ = self.run_cmd(
            [
                'iam',
                'list-users',
                '--output',
                'yaml',
                '--query',
                'Users[1].UserId',
            ],
            expected_rc=0,
        )
        self.assertIn('\u2713', ensure_text_type(stdout))



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.autoprompt.filters import (
    doc_section_visible,
    doc_window_has_focus,
    help_section_visible,
    input_buffer_has_focus,
    is_history_mode,
    is_multi_column,
    is_one_column,
    search_input_has_focus,
)
from awscli.testutils import mock, unittest


class TestFilters(unittest.TestCase):
    @mock.patch('awscli.autoprompt.filters.get_app')
    def test_search_input_has_focus(self, get_app):
        app = mock.Mock()
        app.layout.current_window.style = 'class:search-toolbar'
        get_app.return_value = app
        self.assertTrue(search_input_has_focus())

    @mock.patch('awscli.autoprompt.filters.get_app')
    def test_help_section_visible(self, get_app):
        app = mock.Mock()
        app.show_help = True
        get_app.return_value = app
        self.assertTrue(help_section_visible())

    @mock.patch('awscli.autoprompt.filters.get_app')
    def test_doc_window_has_focus(self, get_app):
        app = mock.Mock()
        app.current_buffer.name = 'doc_buffer'
        get_app.return_value = app
        self.assertTrue(doc_window_has_focus())

    @mock.patch('awscli.autoprompt.filters.get_app')
    def test_doc_section_visible(self, get_app):
        app = mock.Mock()
        app.show_doc = True
        get_app.return_value = app
        self.assertTrue(doc_section_visible())

    @mock.patch('awscli.autoprompt.filters.get_app')
    def test_is_history_mode(self, get_app):
        buffer = mock.Mock()
        buffer.history_mode = True
        app = mock.Mock()
        app.current_buffer = buffer
        app.current_buffer.name = 'input_buffer'
        get_app.return_value = app
        self.assertTrue(is_history_mode())

    @mock.patch('awscli.autoprompt.filters.get_app')
    def test_is_multi_column(self, get_app):
        app = mock.Mock()
        app.multi_column = True
        get_app.return_value = app
        self.assertTrue(is_multi_column())

    @mock.patch('awscli.autoprompt.filters.get_app')
    def test_is_one_column(self, get_app):
        app = mock.Mock()
        app.multi_column = False
        get_app.return_value = app
        self.assertTrue(is_one_column())

    @mock.patch('awscli.autoprompt.filters.get_app')
    def test_input_buffer_has_focus(self, get_app):
        app = mock.Mock()
        app.current_buffer.name = 'input_buffer'
        get_app.return_value = app
        self.assertTrue(input_buffer_has_focus())



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import shutil

from awscli.testutils import BaseAWSCommandParamsTest, FileCreator


class TestDescribeVolumes(BaseAWSCommandParamsTest):
    prefix = 'ec2 describe-volumes'

    def setUp(self):
        super(TestDescribeVolumes, self).setUp()
        self.file_creator = FileCreator()

    def tearDown(self):
        super(TestDescribeVolumes, self).tearDown()
        shutil.rmtree(self.file_creator.rootdir)

    def test_max_results_set_by_default(self):
        command = self.prefix
        params = {'MaxResults': 1000}
        self.assert_params_for_cmd(command, params)

    def test_max_results_not_set_with_volume_ids(self):
        command = self.prefix + ' --volume-ids id-volume'
        params = {'VolumeIds': ['id-volume']}
        self.assert_params_for_cmd(command, params)

    def test_max_results_not_set_with_filter(self):
        command = self.prefix + ' --filters Name=volume-id,Values=id-volume'
        params = {'Filters': [{'Name': 'volume-id', 'Values': ['id-volume']}]}
        self.assert_params_for_cmd(command, params)

    def test_max_results_not_overwritten(self):
        command = self.prefix + ' --max-results 5'
        params = {'MaxResults': 5}
        self.assert_params_for_cmd(command, params)

        command = self.prefix + ' --page-size 5'
        self.assert_params_for_cmd(command, params)

    def test_max_results_with_cli_input_json(self):
        params = {'VolumeIds': ['vol-12345']}
        file_path = self.file_creator.create_file(
            'params.json', json.dumps(params)
        )

        command = self.prefix + ' --cli-input-json file://%s' % file_path
        self.assert_params_for_cmd(command, params)



import os
import re
import sys
from pathlib import Path
from subprocess import run

import pytest

ROOT = Path(__file__).parents[4]
REQUIREMENTS_PATH = ROOT / "requirements"
REGENERATE_LOCK_FILE_SCRIPT_PATH = ROOT / "scripts" / "regenerate-lock-files"

# Lockfiles are generated on the CANNONICAL_PYTHON_VERSION
# These tests will be skipped on any other version of python.
CANNONICAL_PYTHON_VERSION = "3.14"
IS_CANNONICAL_PYTHON_VERSION = sys.version_info[0:2] == tuple(
    map(int, CANNONICAL_PYTHON_VERSION.split("."))
)
SKIP_REASON = f"Lock files are generated on Python {CANNONICAL_PYTHON_VERSION}"

IS_LINUX = sys.platform == "linux"


def should_read_line(line):
    line = line.strip()
    return not line.startswith("#")


def read_lock_file(path):
    with open(path) as f:
        lines = f.read().strip('\n').split('\n')

    # Find source files
    source_files = get_source_files(lines)
    dependencies = []
    for source_file in source_files:
        dependencies.extend(get_dependency_names(source_file))

    # Filter out comments
    lines = [line for line in lines if should_read_line(line)]

    # Filter out transient dependencies, we only care about the ones explicitly
    # mentioned in the requirements files.
    reading = False
    relevant_lines = []
    for line in lines:
        if line and line[0] != ' ':
            dependency_name = get_name_component(line)
            if dependency_name in dependencies:
                reading = True
            else:
                reading = False
        if reading:
            relevant_lines.append(line)
    lockfile = "\n".join(relevant_lines)
    return lockfile


def get_name_component(dependency):
    return re.split(r'(>|<|>=|<=|==)', dependency)[0]


def get_requires_from_pyproject():
    dependency_block_re = re.compile(
        r"dependencies = \[([\s\S]+?)\]\s", re.MULTILINE
    )
    extract_dependencies_re = re.compile(r'"(.+)"')
    with open(ROOT / "pyproject.toml") as f:
        data = f.read()
    raw_dependencies = dependency_block_re.findall(data)[0]
    dependencies = extract_dependencies_re.findall(raw_dependencies)
    return dependencies


def get_dependency_names(filename):
    if not filename.endswith('.txt'):
        return [
            get_name_component(dep) for dep in get_requires_from_pyproject()
        ]

    dependencies = []
    filepath = ROOT / filename
    with open(filepath) as f:
        content = f.read()
        lines = content.split('\n')
        for line in lines:
            if not line:
                continue
            if line.startswith('-r'):
                directory = os.path.dirname(filepath)
                new_filename = line.split(' ')[1]
                new_full_filepath = os.path.abspath(
                    os.path.join(directory, new_filename)
                )
                dependencies.extend(get_dependency_names(new_full_filepath))
            else:
                dependencies.append(get_name_component(line))
    return dependencies


def get_source_files(lines):
    for line in lines:
        if 'pip-compile' and 'generate-hashes' in line:
            line = line[1:]
            files = [
                part
                for part in line.split(' ')
                if part and not part.startswith('--') and part != 'pip-compile'
            ]
            return files


def is_lockfile(path: str) -> bool:
    return path.endswith("-lock.txt")


def lockfile_paths(root):
    for base, _, filenames in os.walk(root):
        for filename in filenames:
            path = os.path.join(base, filename)
            if is_lockfile(path):
                yield path


@pytest.mark.skipif(
    not IS_CANNONICAL_PYTHON_VERSION,
    reason=SKIP_REASON,
)
def test_all_lock_files_are_generated_by_expected_python_version():
    for path in lockfile_paths(REQUIREMENTS_PATH):
        with open(path) as f:
            content = f.read()
            assert f"Python {CANNONICAL_PYTHON_VERSION}" in content


@pytest.mark.skipif(
    not IS_CANNONICAL_PYTHON_VERSION,
    reason=SKIP_REASON,
)
@pytest.mark.skipif(
    IS_LINUX,
    reason=(
        "The linux lock files are generated on mac. So the only "
        "platforms we need to test are mac and Windows."
    ),
)
def test_lock_files_are_up_to_date(tmpdir):
    reqs_dir = tmpdir / "requirements"
    reqs_dir.mkdir()
    download_deps_dir = reqs_dir / "download-deps"
    download_deps_dir.mkdir()
    command = [
        sys.executable,
        REGENERATE_LOCK_FILE_SCRIPT_PATH,
        "--output-directory",
        tmpdir,
    ]
    result = run(command, cwd=ROOT)
    assert result.returncode == 0

    lockfile_mapping = {
        path: path.replace(str(tmpdir), str(ROOT))
        for path in lockfile_paths(tmpdir)
    }

    for regenerated_file, original_file in lockfile_mapping.items():
        assert read_lock_file(regenerated_file) == read_lock_file(
            original_file
        )



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestDeregisterInstancesFromLoadBalancer(BaseAWSCommandParamsTest):
    prefix = 'elb deregister-instances-from-load-balancer'

    def test_shorthand(self):
        command = self.prefix + ' --load-balancer-name foo'
        command += ' --instances id1 id2 id3'
        expected_params = {
            'LoadBalancerName': 'foo',
            'Instances': [
                {'InstanceId': 'id1'},
                {'InstanceId': 'id2'},
                {'InstanceId': 'id3'},
            ],
        }
        self.assert_params_for_cmd(command, expected_params)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json

from awscli.testutils import BaseAWSCommandParamsTest, mock


class TestMonitorMutatingGatewayServiceIntegration(BaseAWSCommandParamsTest):
    def setUp(self):
        super().setUp()
        self.patch_make_request()

    def test_create_gateway_service_with_monitoring(self):
        # Setup mock responses: only valid create response, no describe response
        self.parsed_responses = [
            # CreateExpressGatewayService response
            {
                "service": {
                    "serviceArn": "arn:aws:ecs:us-east-1:123456789:service/test-cluster/test-service",
                    "serviceName": "test-service",
                    "status": "ACTIVE",
                }
            }
            # No DescribeExpressGatewayService response - will cause an error and monitoring to terminate
        ]

        # Execute command with monitoring enabled
        cmdline = 'ecs create-express-gateway-service --service-name test-service --cluster test-cluster --execution-role-arn arn:aws:iam::123456789:role/exec --infrastructure-role-arn arn:aws:iam::123456789:role/infra --primary-container image=nginx --monitor-resources'

        self.run_cmd(cmdline, expected_rc=0)

        # Verify API calls were made (create + monitoring attempted)
        operation_names = [op[0].name for op in self.operations_called]
        assert 'CreateExpressGatewayService' in operation_names
        # Note: DescribeExpressGatewayService calls happen via after-call events
        # which are not supported by the functional test framework

    def test_create_gateway_service_without_monitoring(self):
        # Setup mock response
        self.parsed_responses = [
            {
                "service": {
                    "serviceArn": "arn:aws:ecs:us-east-1:123456789:service/test-cluster/test-service",
                    "serviceName": "test-service",
                    "clusterArn": "arn:aws:ecs:us-east-1:123456789:cluster/test-cluster",
                    "status": "ACTIVE",
                }
            }
        ]

        # Execute command without monitoring flag (default behavior)
        cmdline = 'ecs create-express-gateway-service --service-name test-service --cluster test-cluster --execution-role-arn arn:aws:iam::123456789:role/exec --infrastructure-role-arn arn:aws:iam::123456789:role/infra --primary-container image=nginx'
        self.run_cmd(cmdline, expected_rc=0)

        # Verify only create operation was called (no monitoring)
        operation_names = [op[0].name for op in self.operations_called]
        assert 'CreateExpressGatewayService' in operation_names

    def test_update_gateway_service_with_monitoring(self):
        # Setup mock responses: only valid update response, no describe response
        self.parsed_responses = [
            # UpdateExpressGatewayService response
            {
                "service": {
                    "serviceArn": "arn:aws:ecs:us-east-1:123456789:service/test-cluster/test-service",
                    "serviceName": "test-service",
                    "status": "ACTIVE",
                }
            }
            # No DescribeExpressGatewayService response - will cause an error and monitoring to terminate
        ]

        # Execute command with monitoring enabled
        cmdline = 'ecs update-express-gateway-service --service-arn arn:aws:ecs:us-east-1:123456789:service/test-service --monitor-resources'

        self.run_cmd(cmdline, expected_rc=0)

        # Verify API calls were made (update + monitoring)
        operation_names = [op[0].name for op in self.operations_called]
        assert 'UpdateExpressGatewayService' in operation_names
        # Note: DescribeExpressGatewayService calls happen via after-call events
        # which are not supported by the functional test framework

    def test_delete_gateway_service_with_monitoring(self):
        # Setup mock responses: only valid delete response, no describe response
        self.parsed_responses = [
            # DeleteExpressGatewayService response
            {
                "service": {
                    "serviceArn": "arn:aws:ecs:us-east-1:123456789:service/test-cluster/test-service",
                    "serviceName": "test-service",
                    "status": "DRAINING",
                }
            }
            # No DescribeExpressGatewayService response - will cause an error and monitoring to terminate
        ]

        # Execute command with monitoring enabled
        cmdline = 'ecs delete-express-gateway-service --service-arn arn:aws:ecs:us-east-1:123456789:service/test-service --monitor-resources'

        self.run_cmd(cmdline, expected_rc=0)

        # Verify API calls were made (delete + monitoring)
        operation_names = [op[0].name for op in self.operations_called]
        assert 'DeleteExpressGatewayService' in operation_names
        # Note: DescribeExpressGatewayService calls happen via after-call events
        # which are not supported by the functional test framework

    def test_api_parameters_not_modified(self):
        # Setup mock response
        self.parsed_responses = [
            {
                "service": {
                    "serviceArn": "arn:aws:ecs:us-east-1:123456789:service/test-cluster/test-service",
                    "serviceName": "test-service",
                }
            }
        ]

        # Execute command with skip monitoring
        cmdline = 'ecs create-express-gateway-service --service-name test-service --cluster test-cluster --execution-role-arn arn:aws:iam::123456789:role/exec --infrastructure-role-arn arn:aws:iam::123456789:role/infra --primary-container image=nginx'
        self.run_cmd(cmdline, expected_rc=0)

        # Verify that monitor-resources parameter is not sent to the API
        request_body = json.loads(self.last_params)
        assert 'monitorResources' not in request_body
        assert 'monitor-resources' not in request_body

        # Verify normal parameters are present
        assert 'serviceName' in request_body
        assert 'cluster' in request_body

    def test_update_gateway_service_without_monitoring(self):
        # Setup mock response
        self.parsed_responses = [
            {
                "service": {
                    "serviceArn": "arn:aws:ecs:us-east-1:123456789:service/test-cluster/test-service",
                    "serviceName": "test-service",
                    "status": "ACTIVE",
                }
            }
        ]

        # Execute command without monitoring flag (default behavior)
        cmdline = 'ecs update-express-gateway-service --service-arn arn:aws:ecs:us-east-1:123456789:service/test-service'
        self.run_cmd(cmdline, expected_rc=0)

        # Verify only update operation was called (no monitoring)
        operation_names = [op[0].name for op in self.operations_called]
        assert 'UpdateExpressGatewayService' in operation_names

    def test_delete_gateway_service_without_monitoring(self):
        # Setup mock response
        self.parsed_responses = [
            {
                "service": {
                    "serviceArn": "arn:aws:ecs:us-east-1:123456789:service/test-cluster/test-service",
                    "serviceName": "test-service",
                    "status": "DRAINING",
                }
            }
        ]

        # Execute command without monitoring flag (default behavior)
        cmdline = 'ecs delete-express-gateway-service --service-arn arn:aws:ecs:us-east-1:123456789:service/test-service'
        self.run_cmd(cmdline, expected_rc=0)

        # Verify only delete operation was called (no monitoring)
        operation_names = [op[0].name for op in self.operations_called]
        assert 'DeleteExpressGatewayService' in operation_names



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import platform
import socket
import struct
import sys
from io import BytesIO, TextIOWrapper
from queue import Queue
from threading import Event
from unittest import mock
from unittest.mock import patch
from urllib.parse import urlparse

import pytest
from awscrt import websocket
from awscrt.http import (
    HttpHeaders,
    HttpProxyAuthenticationType,
    HttpProxyOptions,
    HttpRequest,
)
from awscrt.io import ClientTlsContext, TlsConnectionOptions, TlsContextOptions
from awscrt.websocket import (
    IncomingFrame,
    OnConnectionSetupData,
    OnConnectionShutdownData,
    OnIncomingFrameCompleteData,
    OnIncomingFramePayloadData,
    OnSendFrameCompleteData,
    Opcode,
    WebSocket,
)

from awscli.customizations.ec2instanceconnect.websocket import (
    BaseWebsocketIO,
    InputClosedError,
    StdinStdoutIO,
    TCPSocketIO,
    Websocket,
    WebsocketException,
    WebsocketManager,
    WindowsStdinStdoutIO,
)
from tests.markers import if_windows, skip_if_windows


class TestWebsocketIO:
    def test_stdin_stdout_io_read(self, monkeypatch):
        io = StdinStdoutIO()

        expected_data = b"Test"
        monkeypatch.setattr('sys.stdin', TextIOWrapper(BytesIO(expected_data)))

        returned_data = io.read(100)

        assert returned_data == expected_data

    def test_stdin_stdout_io_write(self, monkeypatch):
        io = StdinStdoutIO()

        mock_text_io_wrapper = mock.Mock(spec=TextIOWrapper)
        monkeypatch.setattr('sys.stdout', mock_text_io_wrapper)

        expected_data = b"Test"
        io.write(expected_data)

        mock_text_io_wrapper.buffer.write.assert_called_with(expected_data)
        assert mock_text_io_wrapper.flush.called

    @skip_if_windows
    @mock.patch("select.select")
    def test_stdin_stdout_io_has_data_to_read_returns_false_non_windows(
        self, mock_select
    ):
        io = StdinStdoutIO()

        expected_socket_list = [sys.stdin]

        mock_select.return_value = None, None, None

        return_val = io.has_data_to_read()

        mock_select.assert_called_with(expected_socket_list, [], [], 0.05)
        assert not return_val

    @skip_if_windows
    @mock.patch("select.select")
    def test_stdin_stdout_io_has_data_to_read_returns_true_non_windows(
        self, mock_select
    ):
        io = StdinStdoutIO()

        expected_socket_list = [sys.stdin]

        mock_select.return_value = True, None, None

        return_val = io.has_data_to_read()

        mock_select.assert_called_with(expected_socket_list, [], [], 0.05)
        assert return_val

    @if_windows
    def test_windows_stdin_stdout_io_has_data_to_read_always_returns_true(
        self,
    ):
        io = WindowsStdinStdoutIO()
        assert io.has_data_to_read()

    def test_tcp_socket_io_read(self):
        mock_conn = mock.Mock(spec=socket.SocketType)
        io = TCPSocketIO(mock_conn)

        expected_data = b"Test"
        mock_conn.recv.return_value = expected_data

        bytes_to_read = 100
        returned_data = io.read(bytes_to_read)

        assert returned_data == expected_data
        mock_conn.recv.assert_called_with(bytes_to_read)

    def test_tcp_socket_io_write(self):
        mock_conn = mock.Mock(spec=socket.SocketType)
        io = TCPSocketIO(mock_conn)

        expected_data = b"Test"

        io.write(expected_data)

        mock_conn.sendall.assert_called_with(expected_data)

    def test_tcp_socket_io_read_raise_error_when_empty_data(self):
        mock_conn = mock.Mock(spec=socket.SocketType)
        io = TCPSocketIO(mock_conn)

        expected_data = b""
        mock_conn.recv.return_value = expected_data

        bytes_to_read = 100
        with pytest.raises(InputClosedError):
            io.read(bytes_to_read)

        mock_conn.recv.assert_called_with(bytes_to_read)


@pytest.fixture
def mock_websocketio():
    return mock.Mock(spec=BaseWebsocketIO)


@pytest.fixture
def mock_crt_websocket():
    return mock.Mock(spec=WebSocket)


@pytest.fixture
def mock_on_connection_event():
    return mock.Mock(spec=Event)


@pytest.fixture
def mock_send_frame_results_queue():
    return mock.Mock(spec=Queue)


@pytest.fixture
def mock_shutdown_event():
    return mock.Mock(spec=Event)


@pytest.fixture
def mock_tls_connection_options():
    return mock.Mock(spec=TlsConnectionOptions)


@pytest.fixture
def mock_handshake_request():
    return mock.Mock(spec=HttpRequest)


@pytest.fixture
def web_socket(
    mock_websocketio,
    mock_crt_websocket,
    mock_on_connection_event,
    mock_send_frame_results_queue,
    mock_shutdown_event,
    mock_tls_connection_options,
):
    web_socket = Websocket(
        mock_websocketio,
        "websocket-id",
        mock_tls_connection_options,
        mock_on_connection_event,
        mock_shutdown_event,
        mock_send_frame_results_queue,
    )
    web_socket._websocket = mock_crt_websocket
    return web_socket


def assert_websocket_closed(
    mock_shutdown_event, mock_crt_websocket, mock_websocketio
):
    assert mock_shutdown_event.set.called
    assert mock_crt_websocket.close.called
    assert mock_websocketio.close.called


class CaptureCRTWebsocket:
    def __init__(self):
        self.on_connection_setup_handler = None
        self.on_connection_shutdown_handler = None
        self.on_incoming_frame_payload_handler = None
        self.on_incoming_frame_complete_handler = None

    def connect(self, **kwargs):
        self.on_connection_setup_handler = kwargs['on_connection_setup']
        self.on_connection_shutdown_handler = kwargs['on_connection_shutdown']
        self.on_incoming_frame_payload_handler = kwargs[
            'on_incoming_frame_payload'
        ]
        self.on_incoming_frame_complete_handler = kwargs[
            'on_incoming_frame_complete'
        ]


@pytest.fixture
def capture_crt_websocket():
    return CaptureCRTWebsocket()


@pytest.fixture
def websocket_url():
    return 'wss://amazon.com/openTunnel?params'


class TestWebsocket:
    @pytest.fixture(autouse=True)
    def patch_crt_websocket(self, monkeypatch, capture_crt_websocket):
        monkeypatch.setattr(
            "awscrt.websocket.connect", capture_crt_websocket.connect
        )

    @patch.object(websocket, "connect")
    @patch.object(websocket, "create_handshake_request")
    def test_connect(
        self,
        mock_handshake,
        mock_connect,
        mock_handshake_request,
        mock_on_connection_event,
        mock_tls_connection_options,
        web_socket,
        websocket_url,
    ):
        parsed_url = urlparse(websocket_url)
        host = parsed_url.hostname
        path = parsed_url.path + "?" + parsed_url.query
        mock_handshake.return_value = mock_handshake_request

        http_headers = HttpHeaders()
        mock_handshake_request.headers = http_headers

        web_socket.connect(websocket_url, 'defined-user-agent')

        mock_handshake.assert_called_with(host=host, path=path)
        mock_tls_connection_options.set_server_name.assert_called_with(host)
        assert 'defined-user-agent' == http_headers.get('User-Agent')
        mock_connect.assert_called_with(
            host=host,
            handshake_request=mock_handshake_request,
            port=443,
            tls_connection_options=mock_tls_connection_options,
            proxy_options=None,
            on_connection_setup=mock.ANY,
            on_connection_shutdown=mock.ANY,
            on_incoming_frame_payload=mock.ANY,
            on_incoming_frame_complete=mock.ANY,
        )

        assert mock_on_connection_event.wait.called

    @patch.object(websocket, "connect")
    @patch.object(websocket, "create_handshake_request")
    @mock.patch.dict(
        os.environ,
        {
            "HTTPS_PROXY": "http://user1:pass1@localhost:8989",
            "NO_PROXY": "different-domain",
        },
    )
    def test_connect_with_proxy(
        self,
        mock_handshake,
        mock_connect,
        mock_on_connection_event,
        mock_tls_connection_options,
        web_socket,
        websocket_url,
    ):
        parsed_url = urlparse(websocket_url)
        host = parsed_url.hostname
        path = parsed_url.path + "?" + parsed_url.query
        mock_handshake.return_value = mock_handshake_request

        web_socket.connect(websocket_url)

        mock_handshake.assert_called_with(host=host, path=path)
        mock_tls_connection_options.set_server_name.assert_called_with(host)
        mock_connect.assert_called_with(
            host=host,
            handshake_request=mock_handshake_request,
            port=443,
            tls_connection_options=mock_tls_connection_options,
            proxy_options=mock.ANY,
            on_connection_setup=mock.ANY,
            on_connection_shutdown=mock.ANY,
            on_incoming_frame_payload=mock.ANY,
            on_incoming_frame_complete=mock.ANY,
        )
        assert (
            "localhost"
            == mock_connect.call_args.kwargs['proxy_options'].host_name
        )
        assert 8989 == mock_connect.call_args.kwargs['proxy_options'].port
        assert (
            HttpProxyAuthenticationType.Basic
            == mock_connect.call_args.kwargs['proxy_options'].auth_type
        )
        assert (
            "user1"
            == mock_connect.call_args.kwargs['proxy_options'].auth_username
        )
        assert (
            "pass1"
            == mock_connect.call_args.kwargs['proxy_options'].auth_password
        )

        assert mock_on_connection_event.wait.called

    @patch.object(websocket, "connect")
    @patch.object(websocket, "create_handshake_request")
    @mock.patch.dict(
        os.environ, {"HTTPS_PROXY": "http://user1:pass1@localhost:8989"}
    )
    def test_connect_with_proxy_but_no_proxy_env_empty(
        self,
        mock_handshake,
        mock_connect,
        mock_on_connection_event,
        mock_tls_connection_options,
        web_socket,
        websocket_url,
    ):
        parsed_url = urlparse(websocket_url)
        host = parsed_url.hostname
        path = parsed_url.path + "?" + parsed_url.query
        mock_handshake.return_value = mock_handshake_request

        web_socket.connect(websocket_url)

        mock_handshake.assert_called_with(host=host, path=path)
        mock_tls_connection_options.set_server_name.assert_called_with(host)
        mock_connect.assert_called_with(
            host=host,
            handshake_request=mock_handshake_request,
            port=443,
            tls_connection_options=mock_tls_connection_options,
            proxy_options=mock.ANY,
            on_connection_setup=mock.ANY,
            on_connection_shutdown=mock.ANY,
            on_incoming_frame_payload=mock.ANY,
            on_incoming_frame_complete=mock.ANY,
        )
        assert (
            "localhost"
            == mock_connect.call_args.kwargs['proxy_options'].host_name
        )
        assert 8989 == mock_connect.call_args.kwargs['proxy_options'].port
        assert (
            HttpProxyAuthenticationType.Basic
            == mock_connect.call_args.kwargs['proxy_options'].auth_type
        )
        assert (
            "user1"
            == mock_connect.call_args.kwargs['proxy_options'].auth_username
        )
        assert (
            "pass1"
            == mock_connect.call_args.kwargs['proxy_options'].auth_password
        )

        assert mock_on_connection_event.wait.called

    @patch.object(websocket, "connect")
    @patch.object(websocket, "create_handshake_request")
    @mock.patch.dict(
        os.environ,
        {
            "HTTPS_PROXY": "http://user1:pass1@localhost:8989",
            "NO_PROXY": "1eice-123.eice.amazon.com",
        },
    )
    def test_connect_with_proxy_define_and_no_proxy_defined(
        self,
        mock_handshake,
        mock_connect,
        mock_handshake_request,
        mock_on_connection_event,
        mock_tls_connection_options,
        web_socket,
    ):
        url = "wss://eice-123.eice.amazon.com/openTunnel?params"
        parsed_url = urlparse(url)
        host = parsed_url.hostname
        path = parsed_url.path + "?" + parsed_url.query
        mock_handshake.return_value = mock_handshake_request

        web_socket.connect(url)

        mock_handshake.assert_called_with(host=host, path=path)
        mock_tls_connection_options.set_server_name.assert_called_with(host)
        mock_connect.assert_called_with(
            host=host,
            handshake_request=mock_handshake_request,
            port=443,
            tls_connection_options=mock_tls_connection_options,
            proxy_options=None,
            on_connection_setup=mock.ANY,
            on_connection_shutdown=mock.ANY,
            on_incoming_frame_payload=mock.ANY,
            on_incoming_frame_complete=mock.ANY,
        )

        assert mock_on_connection_event.wait.called

    def test_write_data_from_input_when_there_is_data_to_read(
        self,
        mock_websocketio,
        mock_crt_websocket,
        mock_send_frame_results_queue,
        mock_shutdown_event,
        web_socket,
        capture_crt_websocket,
    ):
        data_as_bytes = b"Test"
        mock_websocketio.read.return_value = data_as_bytes
        mock_websocketio.has_data_to_read.return_value = True

        mock_shutdown_event.is_set.side_effect = [False, True]

        web_socket.write_data_from_input()

        mock_websocketio.read.assert_called_with(65_000)
        mock_crt_websocket.send_frame.assert_called_with(
            opcode=Opcode.BINARY,
            payload=data_as_bytes,
            on_complete=mock.ANY,
        )
        assert mock_send_frame_results_queue.get.called
        assert_websocket_closed(
            mock_shutdown_event, mock_crt_websocket, mock_websocketio
        )

    @mock.patch("time.sleep")
    def test_write_data_from_input_when_there_is_no_data_to_read(
        self,
        mock_sleep,
        mock_websocketio,
        mock_crt_websocket,
        mock_shutdown_event,
        web_socket,
    ):
        mock_websocketio.has_data_to_read.return_value = False

        mock_shutdown_event.is_set.side_effect = [False, True]

        web_socket.write_data_from_input()

        mock_sleep.assert_called_with(0.05)

        assert_websocket_closed(
            mock_shutdown_event, mock_crt_websocket, mock_websocketio
        )

    def test_write_data_from_input_shuts_down_when_sends_frame_throws_close_frame_sent(
        self,
        mock_websocketio,
        mock_crt_websocket,
        mock_shutdown_event,
        web_socket,
    ):
        mock_crt_websocket.send_frame.side_effect = RuntimeError(
            "AWS_ERROR_HTTP_WEBSOCKET_CLOSE_FRAME_SENT"
        )

        data_as_bytes = b"Test"
        mock_websocketio.read.return_value = data_as_bytes
        mock_websocketio.has_data_to_read.return_value = True
        mock_shutdown_event.is_set.side_effect = [False, True]

        web_socket.write_data_from_input()

        assert_websocket_closed(
            mock_shutdown_event, mock_crt_websocket, mock_websocketio
        )

    def test_write_data_northbound_shuts_down_when_sends_frame_throws_http_connection_closed(
        self,
        mock_websocketio,
        mock_crt_websocket,
        mock_shutdown_event,
        web_socket,
    ):
        mock_crt_websocket.send_frame.side_effect = RuntimeError(
            "AWS_ERROR_HTTP_CONNECTION_CLOSED"
        )

        data_as_bytes = b"Test"
        mock_websocketio.read.return_value = data_as_bytes
        mock_websocketio.has_data_to_read.return_value = True
        mock_shutdown_event.is_set.side_effect = [False, True]

        web_socket.write_data_from_input()

        assert_websocket_closed(
            mock_shutdown_event, mock_crt_websocket, mock_websocketio
        )

    def test_on_connection_sets_websocket(
        self,
        capture_crt_websocket,
        websocket_url,
        mock_on_connection_event,
        web_socket,
    ):
        data = OnConnectionSetupData()
        data.websocket = mock.Mock(spec=Websocket)
        request_id = "request_id"
        data.handshake_response_headers = [("X-Amzn-RequestId", request_id)]

        web_socket.connect(websocket_url)
        capture_crt_websocket.on_connection_setup_handler(data)

        assert data.websocket == web_socket._websocket
        assert mock_on_connection_event.set.called

    def test_on_connection_with_exception(
        self,
        mock_on_connection_event,
        web_socket,
        websocket_url,
        capture_crt_websocket,
    ):
        data = OnConnectionSetupData()
        request_id = "request_id"
        data.handshake_response_headers = [("X-Amzn-RequestId", request_id)]
        data.exception = Exception()
        data.handshake_response_status = 401
        data.handshake_response_body = b"Failed to connect"

        web_socket.connect(websocket_url)
        capture_crt_websocket.on_connection_setup_handler(data)

        assert mock_on_connection_event.set.called

    def test_on_incoming_frame_payload_data_with_binary_frame(
        self,
        mock_websocketio,
        web_socket,
        websocket_url,
        capture_crt_websocket,
    ):
        data_in_bytes = b"Test"
        frame = IncomingFrame(
            opcode=Opcode.BINARY, payload_length=len(data_in_bytes), fin=True
        )

        web_socket.connect(websocket_url)
        capture_crt_websocket.on_incoming_frame_payload_handler(
            OnIncomingFramePayloadData(data=data_in_bytes, frame=frame)
        )

        mock_websocketio.write.assert_called_with(data_in_bytes)

    def test_on_incoming_frame_payload_data_with_close_frame(
        self, web_socket, websocket_url, capture_crt_websocket
    ):
        data_in_bytes = b"Test"
        frame = IncomingFrame(
            opcode=Opcode.CLOSE, payload_length=len(data_in_bytes), fin=True
        )

        web_socket.connect(websocket_url)
        capture_crt_websocket.on_incoming_frame_payload_handler(
            OnIncomingFramePayloadData(data=data_in_bytes, frame=frame)
        )

        assert bytearray(data_in_bytes) == web_socket._close_frame_bytes

    def test_on_incoming_frame_payload_data_with_text_frame_sets_shutdown_event(
        self,
        mock_websocketio,
        mock_crt_websocket,
        mock_shutdown_event,
        web_socket,
        websocket_url,
        capture_crt_websocket,
    ):
        data_in_bytes = b"Test"
        frame = IncomingFrame(
            opcode=Opcode.TEXT, payload_length=len(data_in_bytes), fin=True
        )

        web_socket.connect(websocket_url)
        capture_crt_websocket.on_incoming_frame_payload_handler(
            OnIncomingFramePayloadData(data=data_in_bytes, frame=frame)
        )

        assert isinstance(web_socket.exception, WebsocketException)

        assert_websocket_closed(
            mock_shutdown_event, mock_crt_websocket, mock_websocketio
        )

    def test_on_incoming_frame_complete_with_close_frame(
        self,
        mock_websocketio,
        mock_crt_websocket,
        mock_shutdown_event,
        web_socket,
        websocket_url,
        capture_crt_websocket,
    ):
        shutdown_reason = b"Test"
        data_in_bytes = struct.pack(">H", 1) + shutdown_reason
        frame = IncomingFrame(
            opcode=Opcode.CLOSE, payload_length=len(shutdown_reason), fin=True
        )

        web_socket.connect(websocket_url)
        capture_crt_websocket.on_incoming_frame_payload_handler(
            OnIncomingFramePayloadData(
                frame=frame,
                data=data_in_bytes,
            )
        )
        capture_crt_websocket.on_incoming_frame_complete_handler(
            OnIncomingFrameCompleteData(frame=frame, exception=None)
        )

        expected_exc = WebsocketException(
            f"Websocket Closure Reason: {shutdown_reason.decode()}"
        )
        assert isinstance(web_socket.exception, WebsocketException)
        # Ensure shutdown reason is decoded properly. (validate properties on WebsocketException)
        assert web_socket.exception.args == expected_exc.args
        assert_websocket_closed(
            mock_shutdown_event, mock_crt_websocket, mock_websocketio
        )

    def test_on_incoming_frame_complete_with_close_frame_ignores_shutdown_code_1000(
        self,
        mock_websocketio,
        mock_crt_websocket,
        mock_shutdown_event,
        web_socket,
        websocket_url,
        capture_crt_websocket,
    ):
        frame = IncomingFrame(opcode=Opcode.CLOSE, payload_length=0, fin=True)
        incoming_frame_data = OnIncomingFrameCompleteData(frame=frame)

        web_socket.connect(websocket_url)
        capture_crt_websocket.on_incoming_frame_complete_handler(
            incoming_frame_data
        )

        assert not web_socket.exception
        assert_websocket_closed(
            mock_shutdown_event, mock_crt_websocket, mock_websocketio
        )

    def test_on_connection_shutdown(
        self,
        mock_websocketio,
        mock_crt_websocket,
        mock_shutdown_event,
        web_socket,
        websocket_url,
        capture_crt_websocket,
    ):
        data = OnConnectionShutdownData()

        web_socket.connect(websocket_url)
        capture_crt_websocket.on_connection_shutdown_handler(data)

        assert_websocket_closed(
            mock_shutdown_event, mock_crt_websocket, mock_websocketio
        )



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

import awscrt.io

from awscli.logger import (
    disable_crt_logging,
    enable_crt_logging,
    remove_stream_logger,
    set_stream_logger,
)
from awscli.testutils import mock, unittest


class TestLogger(unittest.TestCase):
    def test_can_set_stream_handler(self):
        set_stream_logger('test_stream_logger', logging.DEBUG)
        log = logging.getLogger('test_stream_logger')
        self.assertEqual(
            log.handlers[0].name, 'test_stream_logger_stream_handler'
        )

    def test_keeps_only_one_stream_handler(self):
        set_stream_logger('test_stream_logger', logging.DEBUG)
        set_stream_logger('test_stream_logger', logging.ERROR)
        log = logging.getLogger('test_stream_logger')
        self.assertEqual(len(log.handlers), 1)
        self.assertEqual(
            log.handlers[0].name, 'test_stream_logger_stream_handler'
        )
        self.assertEqual(log.handlers[0].level, logging.ERROR)

    def test_can_remove_stream_handler(self):
        set_stream_logger('test_stream_logger', logging.DEBUG)
        remove_stream_logger('test_stream_logger')
        log = logging.getLogger('test_stream_logger')
        self.assertEqual(len(log.handlers), 0)

    @mock.patch('awscrt.io.set_log_level')
    def test_can_enable_crt_logging(self, mock_init_logging):
        enable_crt_logging()
        mock_init_logging.assert_called_with(
            awscrt.io.LogLevel.Debug,
        )

    @mock.patch('awscrt.io.set_log_level')
    def test_can_disable_crt_logging(self, mock_init_logging):
        disable_crt_logging()
        mock_init_logging.assert_called_with(
            awscrt.io.LogLevel.NoLogs,
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os

from botocore.loaders import Loader
from botocore.session import get_session

from awscli.customizations.configure.addmodel import get_model_location
from awscli.testutils import FileCreator, unittest


class TestGetModelLocation(unittest.TestCase):
    def setUp(self):
        self.session = get_session()
        self.files = FileCreator()

        # Create our own loader for the unit test and not rely on the
        # customer's actual ~/.aws/models nor the builtin botocore data
        # directory.
        self.customer_data_root = os.path.join(self.files.rootdir, 'customer')
        os.mkdir(self.customer_data_root)

        self.builtin_data_root = os.path.join(self.files.rootdir, 'builtin')
        os.mkdir(self.builtin_data_root)

        self.data_loader = Loader(
            [self.customer_data_root, self.builtin_data_root],
            include_default_search_paths=False,
        )
        self.data_loader.CUSTOMER_DATA_PATH = self.customer_data_root
        self.session.register_component('data_loader', self.data_loader)

        # Add some models into the builtin model directory
        # We are going to add two models. One with a matching service name
        # and endpoint and another without.
        self.matching_service = 'matching'
        self.non_matching_service = 'nonmatching'
        self.non_matching_prefix = 'nonmatching-prefix'
        self.default_api_version = '2015-10-01'

        matching_service_path = os.path.join(
            self.builtin_data_root,
            self.matching_service,
            self.default_api_version,
            'service-2.json',
        )
        os.makedirs(os.path.dirname(matching_service_path))

        non_matching_service_path = os.path.join(
            self.builtin_data_root,
            self.non_matching_service,
            self.default_api_version,
            'service-2.json',
        )
        os.makedirs(os.path.dirname(non_matching_service_path))

        # Write the models to the builtin directory
        with open(matching_service_path, 'w') as f:
            json.dump(
                self._create_service_definition(
                    self.matching_service, self.default_api_version
                ),
                f,
            )

        with open(non_matching_service_path, 'w') as f:
            json.dump(
                self._create_service_definition(
                    self.non_matching_prefix, self.default_api_version
                ),
                f,
            )

    def tearDown(self):
        self.files.remove_all()

    def _create_service_definition(self, endpoint_prefix, api_version):
        return {
            "version": "2.0",
            "metadata": {
                "apiVersion": api_version,
                "endpointPrefix": endpoint_prefix,
            },
            "operations": {},
            "shapes": {},
        }

    def test_get_model_location_for_matching_prefix_and_name(self):
        model_location = get_model_location(
            self.session,
            self._create_service_definition(
                self.matching_service, self.default_api_version
            ),
        )
        self.assertEqual(
            os.path.join(
                self.data_loader.CUSTOMER_DATA_PATH,
                self.matching_service,
                self.default_api_version,
                'service-2.json',
            ),
            model_location,
        )

    def test_get_model_location_with_nonmatching_prefix_and_name(self):
        model_location = get_model_location(
            self.session,
            self._create_service_definition(
                self.non_matching_prefix, self.default_api_version
            ),
        )
        self.assertEqual(
            os.path.join(
                self.data_loader.CUSTOMER_DATA_PATH,
                self.non_matching_service,
                self.default_api_version,
                'service-2.json',
            ),
            model_location,
        )

    def test_get_model_location_of_nonexistent_service(self):
        model_location = get_model_location(
            self.session,
            self._create_service_definition(
                'nonexistent', self.default_api_version
            ),
        )
        self.assertEqual(
            os.path.join(
                self.data_loader.CUSTOMER_DATA_PATH,
                'nonexistent',
                self.default_api_version,
                'service-2.json',
            ),
            model_location,
        )

    def test_get_model_location_when_service_name_provided(self):
        model_location = get_model_location(
            self.session,
            self._create_service_definition(
                'nonexistent', self.default_api_version
            ),
            'override',
        )
        self.assertEqual(
            os.path.join(
                self.data_loader.CUSTOMER_DATA_PATH,
                'override',
                self.default_api_version,
                'service-2.json',
            ),
            model_location,
        )

    def test_get_model_location_with_non_v2(self):
        service_definition = self._create_service_definition(
            'existent', self.default_api_version
        )
        service_definition['version'] = '3.0'
        model_location = get_model_location(self.session, service_definition)
        self.assertEqual(
            os.path.join(
                self.data_loader.CUSTOMER_DATA_PATH,
                'existent',
                self.default_api_version,
                'service-3.json',
            ),
            model_location,
        )

    def test_get_model_location_with_missing_version(self):
        service_definition = self._create_service_definition(
            'existent', self.default_api_version
        )
        service_definition.pop('version')
        model_location = get_model_location(self.session, service_definition)
        self.assertEqual(
            os.path.join(
                self.data_loader.CUSTOMER_DATA_PATH,
                'existent',
                self.default_api_version,
                'service-2.json',
            ),
            model_location,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from awscli.compat import StringIO
from awscli.customizations.configure.importer import (
    ConfigureImportCommand,
    CredentialImporter,
    CredentialParserError,
    CSVCredentialParser,
)
from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.testutils import mock, unittest

from . import FakeSession

CSV_HEADERS = (
    'User name,Password,Access key ID,Secret access key,Console login link\n'
)


class TestConfigureImportCommand(unittest.TestCase):
    def setUp(self):
        self.session = FakeSession({'config_file': 'myconfigfile'})
        self.fake_credentials_filename = os.path.expanduser(
            '~/fake_credentials_filename'
        )
        self.session.profile = None
        self.mock_writer = mock.Mock(spec=ConfigFileWriter)
        self.importer = CredentialImporter(self.mock_writer)
        self.stdout = StringIO()
        self.import_command = ConfigureImportCommand(
            self.session,
            importer=self.importer,
            out_stream=self.stdout,
        )

    def _assert_expected_profile(self, args, profile):
        self.import_command(args=args, parsed_globals=None)
        update_args, _ = self.mock_writer.update_config.call_args
        self.assertEqual(update_args[0], profile)
        self.assertIn('/fake_credentials_filename', update_args[1])
        self.assertIn(
            'Successfully imported 1 profile', self.stdout.getvalue()
        )

    def test_import_downloaded_csv(self):
        row = 'PROFILENAME,PW,AKID,SAK,https://console.link\n'
        content = CSV_HEADERS + row
        expected_profile = {
            '__section__': 'PROFILENAME',
            'aws_access_key_id': 'AKID',
            'aws_secret_access_key': 'SAK',
        }
        self._assert_expected_profile(['--csv', content], expected_profile)

    def test_import_downloaded_csv_custom_prefix(self):
        row = 'PROFILENAME,PW,AKID,SAK,https://console.link\n'
        content = CSV_HEADERS + row
        args = ['--csv', content, '--profile-prefix', 'foo-']
        expected_profile = {
            '__section__': 'foo-PROFILENAME',
            'aws_access_key_id': 'AKID',
            'aws_secret_access_key': 'SAK',
        }
        self._assert_expected_profile(args, expected_profile)

    def test_import_downloaded_csv_multiple(self):
        content = (
            'User name,Access key ID,Secret access key\n'
            'PROFILENAME1,AKID1,SAK1\n'
            'PROFILENAME2,AKID2,SAK2\n'
        )
        self.import_command(args=['--csv', content], parsed_globals=None)
        self.assertEqual(self.mock_writer.update_config.call_count, 2)
        self.assertIn(
            'Successfully imported 2 profile', self.stdout.getvalue()
        )

    def test_import_downloaded_bad_headers(self):
        content = 'User name,Secret access key\n'
        with self.assertRaises(CredentialParserError):
            self.import_command(args=['--csv', content], parsed_globals=None)

    def test_raises_error_when_plain_file_path_passed(self):
        with open('temp_creds.csv', 'w') as f:
            f.write('User name,Access key ID,Secret access key\nuser,AKID,SAK')
        try:
            with self.assertRaises(ValueError) as cm:
                self.import_command(
                    args=['--csv', 'temp_creds.csv'], parsed_globals=None
                )
            self.assertIn("without the 'file://' prefix", str(cm.exception))
        finally:
            os.remove('temp_creds.csv')

    def test_inline_csv_succeeds(self):
        csv_string = 'User name,Access key ID,Secret access key\nuser,AKID,SAK'
        self.import_command(args=['--csv', csv_string], parsed_globals=None)
        self.assertIn(
            'Successfully imported 1 profile', self.stdout.getvalue()
        )

    def test_csv_content_from_file_succeeds(self):
        with open('temp_creds.csv', 'w') as f:
            f.write('User name,Access key ID,Secret access key\nuser,AKID,SAK')
        try:
            with open('temp_creds.csv') as f:
                contents = f.read()
            self.import_command(args=['--csv', contents], parsed_globals=None)
            self.assertIn(
                'Successfully imported 1 profile', self.stdout.getvalue()
            )
        finally:
            os.remove('temp_creds.csv')

    @mock.patch('awscli.customizations.configure.importer.warn_if_permissive')
    def test_warn_called_once_when_importing_credentials(self, mock_warn):
        rows = (
            'PROF1,PW,AKID1,SAK1,https://console.link\n'
            'PROF2,PW,AKID2,SAK2,https://console.link\n'
        )
        content = CSV_HEADERS + rows
        self.import_command(args=['--csv', content], parsed_globals=None)
        mock_warn.assert_called_once_with(self.fake_credentials_filename)


class TestCSVCredentialParser(unittest.TestCase):
    def setUp(self):
        self.parser = CSVCredentialParser()
        self.expected_credentials = [('PROFILENAME', 'AKID', 'SAK')]

    def assert_parse_matches_expected(self, contents):
        credentials = self.parser.parse_credentials(contents)
        self.assertEqual(credentials, self.expected_credentials)

    def test_csv_parser_downloaded_csv(self):
        row = 'PROFILENAME,PW,AKID,SAK,https://console.link\n'
        self.assert_parse_matches_expected(CSV_HEADERS + row)

    def test_csv_parser_simple(self):
        contents = (
            'User name,Access key ID,Secret access key\nPROFILENAME,AKID,SAK\n'
        )
        self.assert_parse_matches_expected(contents)

    def test_csv_parser_with_bom(self):
        contents = (
            '\ufeffUser name,Access key ID,Secret access key\n'
            'PROFILENAME,AKID,SAK\n'
        )
        self.assert_parse_matches_expected(contents)

    def test_csv_parser_multiple_entries(self):
        contents = (
            'User name,Access key ID,Secret access key\n'
            'PROFILENAME1,AKID1,SAK1\n'
            'PROFILENAME2,AKID2,SAK2\n'
        )
        self.expected_credentials = [
            ('PROFILENAME1', 'AKID1', 'SAK1'),
            ('PROFILENAME2', 'AKID2', 'SAK2'),
        ]
        self.assert_parse_matches_expected(contents)

    def test_csv_parser_multiple_entries_bom(self):
        contents = (
            '\ufeffUser name,Access key ID,Secret access key\n'
            'PROFILENAME1,AKID1,SAK1\n'
            'PROFILENAME2,AKID2,SAK2\n'
        )
        self.expected_credentials = [
            ('PROFILENAME1', 'AKID1', 'SAK1'),
            ('PROFILENAME2', 'AKID2', 'SAK2'),
        ]
        self.assert_parse_matches_expected(contents)

    def test_csv_parser_multiple_entries_with_bad_entries(self):
        self.parser.strict = False
        contents = (
            'User name,Access key ID,Secret access key\n'
            'PROFILENAME1,AKID1,SAK1\n'
            'PROFILENAME,,SAK\n'
            'PROFILENAME,AKID,\n'
            ',AKID,SAK\n'
            'PROFILENAME2,AKID2,SAK2\n'
        )
        self.expected_credentials = [
            ('PROFILENAME1', 'AKID1', 'SAK1'),
            ('PROFILENAME2', 'AKID2', 'SAK2'),
        ]
        self.assert_parse_matches_expected(contents)

    def test_csv_parser_header_reorder(self):
        contents = (
            'Access key ID,Secret access key, User name\n'
            'AKID,SAK,PROFILENAME\n'
        )
        self.assert_parse_matches_expected(contents)

    def test_csv_parser_header_case_insensitive(self):
        contents = (
            'access key id,secret access key, user name\n'
            'AKID,SAK,PROFILENAME\n'
        )
        self.assert_parse_matches_expected(contents)

    def test_csv_parser_header_whitespace_insensitive(self):
        contents = (
            '  user name,  access key id,  secret access key    \n'
            'PROFILENAME,AKID,SAK,PROFILENAME\n'
        )
        self.assert_parse_matches_expected(contents)

    def test_csv_parser_values_whitespace_insensitive(self):
        contents = (
            'User name,Access key ID,Secret access key\n'
            '   PROFILENAME   , AKID   ,    SAK  \n'
        )
        self.assert_parse_matches_expected(contents)

    def test_csv_parser_empty_contents(self):
        with self.assertRaises(CredentialParserError):
            self.parser.parse_credentials(' ')

    def test_csv_parser_no_username(self):
        row = ',PW,AKID,SAK,https://console.link\n'
        with self.assertRaises(CredentialParserError):
            self.parser.parse_credentials(CSV_HEADERS + row)

    def test_csv_parser_no_username_header(self):
        contents = 'Access key ID,Secret access key\n'
        with self.assertRaises(CredentialParserError):
            self.parser.parse_credentials(contents)

    def test_csv_parser_no_akid(self):
        row = 'PROFILENAME,PW,,SAK,https://console.link\n'
        with self.assertRaises(CredentialParserError):
            self.parser.parse_credentials(CSV_HEADERS + row)

    def test_csv_parser_no_akid_header(self):
        contents = 'User name,Secret access key\n'
        with self.assertRaises(CredentialParserError):
            self.parser.parse_credentials(contents)

    def test_csv_parser_no_sak(self):
        row = 'PROFILENAME,PW,AKID,,https://console.link\n'
        with self.assertRaises(CredentialParserError):
            self.parser.parse_credentials(CSV_HEADERS + row)

    def test_csv_parser_no_sak_header(self):
        contents = 'User name,Access key ID\n'
        with self.assertRaises(CredentialParserError):
            self.parser.parse_credentials(contents)


class TestCredentialImporter(unittest.TestCase):
    def setUp(self):
        self.mock_writer = mock.Mock(spec=ConfigFileWriter)
        self.importer = CredentialImporter(self.mock_writer)

    def test_import_profile(self):
        file = 'credentials_file'
        credential = ('USERNAME', 'AKID', 'SAK')
        self.importer.import_credential(credential, file)
        profile = {
            '__section__': 'USERNAME',
            'aws_access_key_id': 'AKID',
            'aws_secret_access_key': 'SAK',
        }
        self.mock_writer.update_config.assert_called_with(profile, file)

    def test_import_profile_with_prefix(self):
        file = 'credentials_file'
        credential = ('USERNAME', 'AKID', 'SAK')
        self.importer.import_credential(credential, file, profile_prefix='a-')
        profile = {
            '__section__': 'a-USERNAME',
            'aws_access_key_id': 'AKID',
            'aws_secret_access_key': 'SAK',
        }
        self.mock_writer.update_config.assert_called_with(profile, file)



#!/usr/bin/env python
# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestCreateStack(BaseAWSCommandParamsTest):
    prefix = 'cloudformation create-stack'

    def test_basic_create_stack(self):
        cmdline = self.prefix
        cmdline += ' --stack-name test-stack --template-url http://foo'
        result = {'StackName': 'test-stack', 'TemplateURL': 'http://foo'}
        self.assert_params_for_cmd(cmdline, result)

    def test_create_stack_string_params(self):
        cmdline = self.prefix
        cmdline += ' --stack-name test-stack --template-url http://foo'
        cmdline += ' --parameters ParameterKey=foo,ParameterValue=bar'
        cmdline += ' ParameterKey=foo2,ParameterValue=bar2'
        result = {
            'StackName': 'test-stack',
            'TemplateURL': 'http://foo',
            'Parameters': [
                {'ParameterKey': 'foo', 'ParameterValue': 'bar'},
                {'ParameterKey': 'foo2', 'ParameterValue': 'bar2'},
            ],
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_create_stack_for_csv_params_escaping(self):
        # If a template is specified as a comma delimited list,
        # we need to be able to quote the value or escape the comma.
        cmdline = self.prefix
        cmdline += ' --stack-name test-stack --template-url http://foo'
        cmdline += r' --parameters ParameterKey=foo,ParameterValue=one\,two'
        result = {
            'StackName': 'test-stack',
            'TemplateURL': 'http://foo',
            'Parameters': [
                {'ParameterKey': 'foo', 'ParameterValue': 'one,two'}
            ],
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_create_stack_for_csv_with_quoting(self):
        cmdline = self.prefix
        cmdline += ' --stack-name test-stack --template-url http://foo'
        # Note how we're quoting the value of parameter_value.
        cmdline += ' --parameters ParameterKey=foo,ParameterValue="one,two"'
        result = {
            'StackName': 'test-stack',
            'TemplateURL': 'http://foo',
            'Parameters': [
                {'ParameterKey': 'foo', 'ParameterValue': 'one,two'}
            ],
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_can_handle_empty_parameters(self):
        cmdline = self.prefix
        cmdline += ' --stack-name test --parameters --template-url http://foo'
        result = {
            'StackName': 'test',
            'TemplateURL': 'http://foo',
            'Parameters': [],
        }
        self.assert_params_for_cmd(cmdline, result)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from tests.unit.customizations.emr import (
    EMRBaseAWSCommandParamsTest as BaseAWSCommandParamsTest,
)


class TestModifyClusterAttributes(BaseAWSCommandParamsTest):
    prefix = 'emr modify-cluster-attributes'

    def test_visible_to_all(self):
        args = ' --cluster-id j-ABC123456 --visible-to-all-users'
        cmdline = self.prefix + args
        result = {'JobFlowIds': ['j-ABC123456'], 'VisibleToAllUsers': True}
        self.assert_params_for_cmd(cmdline, result)

    def test_no_visible_to_all(self):
        args = ' --cluster-id j-ABC123456 --no-visible-to-all-users'
        cmdline = self.prefix + args
        result = {'JobFlowIds': ['j-ABC123456'], 'VisibleToAllUsers': False}
        self.assert_params_for_cmd(cmdline, result)

    def test_termination_protected(self):
        args = ' --cluster-id j-ABC123456 --termination-protected'
        cmdline = self.prefix + args
        result = {'JobFlowIds': ['j-ABC123456'], 'TerminationProtected': True}
        self.assert_params_for_cmd(cmdline, result)

    def test_no_termination_protected(self):
        args = ' --cluster-id j-ABC123456 --no-termination-protected'
        cmdline = self.prefix + args
        result = {'JobFlowIds': ['j-ABC123456'], 'TerminationProtected': False}
        self.assert_params_for_cmd(cmdline, result)

    def test_no_auto_terminate(self):
        args = ' --cluster-id j-ABC123456 --no-auto-terminate'
        cmdline = self.prefix + args
        result = {
            'JobFlowIds': ['j-ABC123456'],
            'KeepJobFlowAliveWhenNoSteps': True,
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_auto_terminate(self):
        args = ' --cluster-id j-ABC123456 --auto-terminate'
        cmdline = self.prefix + args
        result = {
            'JobFlowIds': ['j-ABC123456'],
            'KeepJobFlowAliveWhenNoSteps': False,
        }

    def test_unhealthy_node_replacement(self):
        args = ' --cluster-id j-ABC123456 --unhealthy-node-replacement'
        cmdline = self.prefix + args
        result = {
            'JobFlowIds': ['j-ABC123456'],
            'UnhealthyNodeReplacement': True,
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_no_unhealthy_node_replacement(self):
        args = ' --cluster-id j-ABC123456 --no-unhealthy-node-replacement'
        cmdline = self.prefix + args
        result = {
            'JobFlowIds': ['j-ABC123456'],
            'UnhealthyNodeReplacement': False,
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_visible_to_all_and_no_visible_to_all(self):
        args = (
            ' --cluster-id j-ABC123456 --no-visible-to-all-users'
            ' --visible-to-all-users'
        )
        cmdline = self.prefix + args
        expected_error_msg = (
            '\naws: [ERROR]: An error occurred (ParamValidation): You cannot specify both --visible-to-all-users '
            'and --no-visible-to-all-users options together.\n'
        )
        result = self.run_cmd(cmdline, 252)
        self.assertEqual(expected_error_msg, result[1])

    def test_temination_protected_and_no_termination_protected(self):
        args = (
            ' --cluster-id j-ABC123456 --no-termination-protected'
            ' --termination-protected'
        )
        cmdline = self.prefix + args
        expected_error_msg = (
            '\naws: [ERROR]: An error occurred (ParamValidation): You cannot specify both --termination-protected '
            'and --no-termination-protected options together.\n'
        )
        result = self.run_cmd(cmdline, 252)
        self.assertEqual(expected_error_msg, result[1])

    def test_auto_terminate_and_no_auto_terminate(self):
        args = (
            ' --cluster-id j-ABC123456 --auto-terminate' ' --no-auto-terminate'
        )
        cmdline = self.prefix + args
        expected_error_msg = (
            '\naws: [ERROR]: An error occurred (ParamValidation): You cannot specify both --auto-terminate '
            'and --no-auto-terminate options together.\n'
        )
        result = self.run_cmd(cmdline, 252)
        self.assertEqual(expected_error_msg, result[1])

    def test_can_set_multiple_attributes(self):
        args = (
            ' --cluster-id j-ABC123456 --termination-protected'
            ' --visible-to-all-users --unhealthy-node-replacement'
        )
        cmdline = self.prefix + args
        result_set_termination_protection = {
            'JobFlowIds': ['j-ABC123456'],
            'TerminationProtected': True,
        }
        result_set_visible_to_all_users = {
            'JobFlowIds': ['j-ABC123456'],
            'VisibleToAllUsers': True,
        }
        result_set_unhealty_node_replacement = {
            'JobFlowIds': ['j-ABC123456'],
            'UnhealthyNodeReplacement': True,
        }
        self.run_cmd(cmdline)
        self.assertDictEqual(
            self.operations_called[0][1], result_set_visible_to_all_users
        )
        self.assertDictEqual(
            self.operations_called[1][1], result_set_termination_protection
        )
        self.assertDictEqual(
            self.operations_called[2][1], result_set_unhealty_node_replacement
        )

    def test_can_set_multiple_attributes_with_no(self):
        args = (
            ' --cluster-id j-ABC123456 --termination-protected'
            ' --no-visible-to-all-users --unhealthy-node-replacement'
        )
        cmdline = self.prefix + args
        result_set_termination_protection = {
            'JobFlowIds': ['j-ABC123456'],
            'TerminationProtected': True,
        }
        result_set_visible_to_all_users = {
            'JobFlowIds': ['j-ABC123456'],
            'VisibleToAllUsers': False,
        }
        result_set_unhealty_node_replacement = {
            'JobFlowIds': ['j-ABC123456'],
            'UnhealthyNodeReplacement': True,
        }
        self.run_cmd(cmdline)
        self.assertDictEqual(
            self.operations_called[0][1], result_set_visible_to_all_users
        )
        self.assertDictEqual(
            self.operations_called[1][1], result_set_termination_protection
        )
        self.assertDictEqual(
            self.operations_called[2][1], result_set_unhealty_node_replacement
        )

    def test_at_least_one_option(self):
        args = ' --cluster-id j-ABC123456'
        cmdline = self.prefix + args
        expected_error_msg = (
            '\naws: [ERROR]: An error occurred (ParamValidation): Must specify one of the following boolean options: '
            '--visible-to-all-users|--no-visible-to-all-users, '
            '--termination-protected|--no-termination-protected, '
            '--auto-terminate|--no-auto-terminate, '
            '--unhealthy-node-replacement|--no-unhealthy-node-replacement.\n'
        )
        result = self.run_cmd(cmdline, 252)
        self.assertEqual(expected_error_msg, result[1])


if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import botocore.session
from botocore.exceptions import ClientError

from awscli.compat import StringIO
from awscli.customizations.configservice.subscribe import (
    S3BucketHelper,
    SNSTopicHelper,
    SubscribeCommand,
)
from awscli.testutils import mock, unittest


class TestS3BucketHelper(unittest.TestCase):
    def setUp(self):
        self.session = botocore.session.get_session()
        self.s3_client = mock.Mock(self.session.create_client('s3'))
        self.helper = S3BucketHelper(self.s3_client)
        self.error_response = {
            'Error': {'Code': '404', 'Message': 'Not Found'}
        }
        self.bucket_no_exists_error = ClientError(
            self.error_response, 'HeadBucket'
        )

    def test_correct_prefix_returned(self):
        name = 'MyBucket/MyPrefix'
        bucket, prefix = self.helper.prepare_bucket(name)
        # Ensure the returned bucket and key are as expected
        self.assertEqual(bucket, 'MyBucket')
        self.assertEqual(prefix, 'MyPrefix')

    def test_bucket_exists(self):
        name = 'MyBucket'
        bucket, prefix = self.helper.prepare_bucket(name)
        # A new bucket should not have been created because no error was thrown
        self.assertFalse(self.s3_client.create_bucket.called)
        # Ensure the returned bucket and key are as expected
        self.assertEqual(bucket, name)
        self.assertEqual(prefix, '')

    def test_bucket_no_exist(self):
        name = 'MyBucket/MyPrefix'
        self.s3_client.head_bucket.side_effect = self.bucket_no_exists_error
        self.s3_client.meta.region_name = 'us-east-1'
        bucket, prefix = self.helper.prepare_bucket(name)
        # Ensure that the create bucket was called with the proper args.
        self.s3_client.create_bucket.assert_called_with(Bucket='MyBucket')
        # Ensure the returned bucket and key are as expected
        self.assertEqual(bucket, 'MyBucket')
        self.assertEqual(prefix, 'MyPrefix')

    def test_bucket_no_exist_with_location_constraint(self):
        name = 'MyBucket/MyPrefix'
        self.s3_client.head_bucket.side_effect = self.bucket_no_exists_error
        self.s3_client.meta.region_name = 'us-west-2'
        bucket, prefix = self.helper.prepare_bucket(name)
        # Ensure that the create bucket was called with the proper args.
        self.s3_client.create_bucket.assert_called_with(
            Bucket='MyBucket',
            CreateBucketConfiguration={'LocationConstraint': 'us-west-2'},
        )
        # Ensure the returned bucket and key are as expected
        self.assertEqual(bucket, 'MyBucket')
        self.assertEqual(prefix, 'MyPrefix')

    def test_bucket_client_exception_non_404(self):
        name = 'MyBucket/MyPrefix'
        self.error_response['Error']['Code'] = '403'
        self.error_response['Error']['Message'] = 'Forbidden'
        forbidden_error = ClientError(self.error_response, 'HeadBucket')
        self.s3_client.head_bucket.side_effect = forbidden_error
        self.s3_client._endpoint.region_name = 'us-east-1'
        bucket, prefix = self.helper.prepare_bucket(name)
        # A new bucket should not have been created because a 404 error
        # was not thrown
        self.assertFalse(self.s3_client.create_bucket.called)
        # Ensure the returned bucket and key are as expected
        self.assertEqual(bucket, 'MyBucket')
        self.assertEqual(prefix, 'MyPrefix')

    def test_output_use_existing_bucket(self):
        name = 'MyBucket/MyPrefix'
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.helper.prepare_bucket(name)
            self.assertIn(
                'Using existing S3 bucket: MyBucket', mock_stdout.getvalue()
            )

    def test_output_create_bucket(self):
        name = 'MyBucket/MyPrefix'
        self.s3_client.head_bucket.side_effect = self.bucket_no_exists_error
        self.s3_client._endpoint.region_name = 'us-east-1'
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.helper.prepare_bucket(name)
            self.assertIn(
                'Using new S3 bucket: MyBucket', mock_stdout.getvalue()
            )


class TestSNSTopicHelper(unittest.TestCase):
    def setUp(self):
        self.session = botocore.session.get_session()
        self.sns_client = mock.Mock(
            self.session.create_client('sns', 'us-east-1')
        )
        self.helper = SNSTopicHelper(self.sns_client)

    def test_sns_topic_by_name(self):
        name = 'mysnstopic'
        self.sns_client.create_topic.return_value = {'TopicArn': 'myARN'}
        sns_arn = self.helper.prepare_topic(name)
        # Ensure that the topic was create and returned the expected arn
        self.assertTrue(self.sns_client.create_topic.called)
        self.assertEqual(sns_arn, 'myARN')

    def test_sns_topic_by_arn(self):
        name = 'arn:aws:sns:us-east-1:934212987125:config'
        sns_arn = self.helper.prepare_topic(name)
        # Ensure that the topic was not created and returned the expected arn
        self.assertFalse(self.sns_client.create_topic.called)
        self.assertEqual(sns_arn, name)

    def test_output_existing_topic(self):
        name = 'mysnstopic'
        self.sns_client.create_topic.return_value = {'TopicArn': 'myARN'}
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.helper.prepare_topic(name)
            self.assertIn('Using new SNS topic: myARN', mock_stdout.getvalue())

    def test_output_new_topic(self):
        name = 'arn:aws:sns:us-east-1:934212987125:config'
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.helper.prepare_topic(name)
            self.assertIn(
                'Using existing SNS topic: %s' % name, mock_stdout.getvalue()
            )


class TestSubscribeCommand(unittest.TestCase):
    def setUp(self):
        self.session = botocore.session.get_session()

        # Set up the client mocks.
        self.s3_client = mock.Mock(self.session.create_client('s3'))
        self.sns_client = mock.Mock(
            self.session.create_client('sns', 'us-east-1')
        )
        self.config_client = mock.Mock(
            self.session.create_client('config', 'us-east-1')
        )
        self.config_client.describe_configuration_recorders.return_value = {
            'ConfigurationRecorders': []
        }
        self.config_client.describe_delivery_channels.return_value = {
            'DeliveryChannels': []
        }

        self.session = mock.Mock(self.session)
        self.session.create_client.side_effect = [
            self.s3_client,
            self.sns_client,
            self.config_client,
        ]

        self.parsed_args = mock.Mock()
        self.parsed_args.s3_bucket = 'MyBucket/MyPrefix'
        self.parsed_args.sns_topic = (
            'arn:aws:sns:us-east-1:934212987125:config'
        )
        self.parsed_args.iam_role = 'arn:aws:iam::1234556789:role/config'

        self.parsed_globals = mock.Mock()

        self.cmd = SubscribeCommand(self.session)

    def test_setup_clients(self):
        self.parsed_globals.verify_ssl = True
        self.parsed_globals.region = 'us-east-1'
        self.parsed_globals.endpoint_url = 'http://myendpoint.com'

        self.cmd._run_main(self.parsed_args, self.parsed_globals)

        # Check to see that the clients were created correctly
        self.session.create_client.assert_any_call(
            's3',
            verify=self.parsed_globals.verify_ssl,
            region_name=self.parsed_globals.region,
        )
        self.session.create_client.assert_any_call(
            'sns',
            verify=self.parsed_globals.verify_ssl,
            region_name=self.parsed_globals.region,
        )
        self.session.create_client.assert_any_call(
            'config',
            verify=self.parsed_globals.verify_ssl,
            region_name=self.parsed_globals.region,
            endpoint_url=self.parsed_globals.endpoint_url,
        )

    def test_subscribe(self):
        self.cmd._run_main(self.parsed_args, self.parsed_globals)

        # Check the call made to put configuration recorder.
        self.config_client.put_configuration_recorder.assert_called_with(
            ConfigurationRecorder={
                'name': 'default',
                'roleARN': self.parsed_args.iam_role,
            }
        )

        # Check the call made to put delivery channel.
        self.config_client.put_delivery_channel.assert_called_with(
            DeliveryChannel={
                'name': 'default',
                's3BucketName': 'MyBucket',
                'snsTopicARN': self.parsed_args.sns_topic,
                's3KeyPrefix': 'MyPrefix',
            }
        )

        # Check the call made to start configuration recorder.
        self.config_client.start_configuration_recorder.assert_called_with(
            ConfigurationRecorderName='default'
        )

        # Check that the describe delivery channel and configuration recorder
        # methods were called.
        self.assertTrue(
            self.config_client.describe_configuration_recorders.called
        )
        self.assertTrue(self.config_client.describe_delivery_channels.called)
